# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Support for Debian(-based) distribution(s)."""
import os
import re

import itertools
from datetime import datetime

import attr
from collections import defaultdict

from six.moves import map

import pytz

from niceman import utils
from niceman.utils import attrib

from email.utils import mktime_tz, parsedate_tz

import logging

import requests

from niceman.support.distributions.debian import \
    parse_apt_cache_show_pkgs_output, parse_apt_cache_policy_pkgs_output, \
    parse_apt_cache_policy_source_info, get_apt_release_file_names, \
    get_spec_from_release_file, parse_dpkgquery_line

# Pick a conservative max command-line
from niceman.utils import get_cmd_batch_len, execute_command_batch, \
    cmd_err_filter, join_sequence_of_dicts

from niceman.distributions.base import DistributionTracer

lgr = logging.getLogger('niceman.distributions.debian')

from .base import SpecObject
from .base import Package
from .base import Distribution
from .base import TypedList
from .base import _register_with_representer
from ..support.exceptions import CommandError
#
# Models
#

# TODO: flyweight/singleton ?
# To make them hashable we need to freeze them... not sure if we are ready:
#@attr.s(cmp=True, hash=True, frozen=True)
@attr.s(cmp=True)
class APTSource(SpecObject):
    """APT origin information
    """
    name = attrib(default=attr.NOTHING)
    component = attrib()
    archive = attrib()
    architecture = attrib()
    codename = attrib()
    origin = attrib()
    label = attrib()
    site = attrib()
    archive_uri = attrib()
    date = attrib()
_register_with_representer(APTSource)


@attr.s(slots=True, frozen=True, cmp=False, hash=True)
class DEBPackage(Package):
    """Debian package information"""
    name = attrib(default=attr.NOTHING)
    # Optional
    upstream_name = attrib()
    version = attrib()
    architecture = attrib()
    source_name = attrib(hash=False)
    source_version = attrib(hash=False)
    size = attrib(hash=False)
    md5 = attrib(hash=False)
    sha1 = attrib(hash=False)
    sha256 = attrib(hash=False)
    versions = attrib(hash=False)  # Hash ver_str -> [Array of source names]
    install_date = attrib(hash=False)
    files = attrib(default=attr.Factory(list), hash=False)

    _cmp_fields = ('name', 'architecture')

    def satisfies(self, other):
        """return True if this package (self) satisfies the requirements of 
        the passed package (other)"""
        if not isinstance(other, Package):
            raise TypeError('satisfies() requires a package argument')
        if not isinstance(other, DEBPackage):
            return False
        if self.name != other.name:
            return False
        if other.version is not None and self.version != other.version:
            return False
        if other.architecture is not None \
                and self.architecture != other.architecture:
            return False
        return True

_register_with_representer(DEBPackage)

@attr.s
class DebianDistribution(Distribution):
    """
    Class to provide Debian-based shell commands.
    """

    apt_sources = TypedList(APTSource)
    packages = TypedList(DEBPackage)
    version = attrib()  # version as depicted by /etc/debian_version

    _cmp_fields = tuple()

    def initiate(self, session):
        """
        Perform any initialization commands needed in the environment.

        Parameters
        ----------
        session : Session object
            The session to work in.
        """
        lgr.debug("Adding Debian update to environment command list.")
        self._init_apt_sources(session)
        session.execute_command(['apt-get', '-o',
            'Acquire::Check-Valid-Until=false', 'update'])
        #session.execute_command(['apt-get', 'install', '-y', 'python-pip'])
        # session.set_env(DEBIAN_FRONTEND='noninteractive', this_session_only=True)

    def _init_apt_sources(self, session,
        apt_source_file='/etc/apt/sources.list.d/niceman.sources.list'):
        """
        Update /etc/apt/sources if necessary based on source date.

        Parameters
        ----------
        session : Session object
        apt_source_file: string
        """

        repo_info = {
            'Debian': {
                'url': 'snapshot.debian.org',
                'keyserver': None,
                'key': None
            },
            'NeuroDebian': {
                'url': 'snapshot-neuro.debian.net:5002',
                'keyserver': 'hkp://pool.sks-keyservers.net:80',
                'key': '0xA5D32F012649A5A9'
            }
        }

        # Create a new apt sources file if needed.
        if not session.exists(apt_source_file):
            session.execute_command(
                "sh -c 'echo \"# Niceman repo sources\" > {}'"
                .format(apt_source_file))

        for source in [s for s in self.apt_sources
            if s.origin in repo_info.keys()]:
            
            # Write snapshot repo to apt sources file.
            date = datetime.strptime(source.date.split('+')[0], "%Y-%m-%d %X")
            template = 'deb http://{}/archive/{}/{}/ {} main contrib non-free'
            source_line = template.format(
                repo_info[source.origin]['url'],
                source.origin.lower(),
                date.strftime("%Y%m%dT%H%M%SZ"),
                source.codename
            )
            self._write_apt_sources(session, apt_source_file, source_line)

            # Write "next" snapshot repo to apt sources file.
            template_list_page = 'http://{}/archive/{}/{}/dists/{}/'
            url = template_list_page.format(
                repo_info[source.origin]['url'],
                source.origin.lower(),
                date.strftime("%Y%m%dT%H%M%SZ"),
                source.codename
            )
            r = requests.get(url)
            m = re.search(
                '<a href="/archive/\w*debian/(\w+)/dists/\w+/">next change</a>',
                r.text)
            if m:
                source_line = template.format(
                    repo_info[source.origin]['url'],
                    source.origin.lower(),
                    m.group(1),
                    source.codename
                )
                self._write_apt_sources(session, apt_source_file, source_line)

            # Add keyserver if needed.
            if repo_info[source.origin]['keyserver']:
                session.execute_command(['apt-key', 'adv', '--recv-keys',
                    '--keyserver', repo_info[source.origin]['keyserver'],
                    repo_info[source.origin]['key']])

    def _write_apt_sources(self, session, apt_source_file, source_line):
        """
        Write a line to the /etc/apt/sources.d/ file

        Parameters
        ----------
        session : Session object
        apt_source_file: string
        source_line: string
        """
        command = "grep -q '{}' {}"
        out, line_not_found = session.execute_command(command.format(
            source_line, apt_source_file))
        if line_not_found:
            lgr.debug("Adding line '{}' to {}".format(source_line,
                apt_source_file))
            session.execute_command("sh -c 'echo {} >> {}'".format(
                source_line, apt_source_file))

    def install_packages(self, session, use_version=True):
        """
        Install the packages associated to this distribution by the provenance
        into the environment.

        Parameters
        ----------
        session : object
        use_version : bool, optional
          Use version information if provided.
          TODO: support outside or deprecate
            
        """
        package_specs = []

        for package in self.packages:
            package_spec = package.name
            if use_version and package.version:
                package_spec += '=%s' % package.version
            package_specs.append(package_spec)

        # Doing in one shot to fail early if any of the versioned specs
        # couldn't be satisfied
        lgr.debug("Installing %s", ', '.join(package_specs))
        session.execute_command(
            # TODO: Pull env out of provenance for this command.
            ['apt-get', 'install', '-y'] + package_specs,
            # env={'DEBIAN_FRONTEND': 'noninteractive'}
        )

    def normalize(self):
        # TODO:
        # - among apt-source we could merge some together if we allow for
        #   e.g. component (main, contrib, non-free) to be a list!  but that
        #   would make us require supporting flexible typing -- string or a list
        pass

    def satisfies_package(self, package):
        """return True if this distribution (self) satisfies the requirements 
        of the passed package"""
        if not isinstance(package, Package):
            raise TypeError('satisfies_package() requires a package argument')
        if not isinstance(package, DEBPackage):
            return False
        return any([ p.satisfies(package) for p in self.packages ])

    def satisfies(self, other):
        """return True if this distribution (self) satisfies the requirements 
        of the other distribution (other)"""
        if not isinstance(other, Distribution):
            raise TypeError('satisfies() requires a distribution argument')
        if not isinstance(other, DebianDistribution):
            return False
        return all(map(self.satisfies_package, other.packages))

    def __sub__(self, other):
        # the semantics of distribution subtraction are, for d1 - d2:
        #     what is specified in d1 that is not specified in d2
        #     or how does d2 fall short of d1
        #     or what is in d1 that isn't satisfied by d2
        return [ p for p in self.packages if not other.satisfies_package(p) ]

    # to grow:
    #  def __iadd__(self, another_instance or DEBPackage, or APTSource)
    #  def __add__(self, another_instance or DEBPackage, or APTSource)
    # but for the checks we should get away from a plain "list of" to some
    # structure to help identify on presence

_register_with_representer(DebianDistribution)


class DebTracer(DistributionTracer):
    """.deb-based (and using apt and dpkg) systems package tracer
    """

    # The Debian tracer is not designed to handle directories
    HANDLES_DIRS = False

    # TODO: (Low Priority) handle cases from dpkg-divert
    def _init(self):
        # TODO: we might want a generic helper for collections of things
        # where we could match based on the set of attrs which matter
        self._apt_sources = {}  # dict of source_name -> APTSource
        self._apt_source_names = set()
        self._all_apt_sources = {}
        self._source_line_to_name_map = {}

    def identify_distributions(self, files):
        if not files:
            return
        
        try:
            debian_version = self._session.read('/etc/debian_version').strip()
            self._session.exists('/etc/os-release')
            # for now would also match Ubuntu -- there it would have
            # ID=ubuntu and ID_LIKE=debian
            # TODO: load/parse /etc/os-release into a dict and better use
            # VERSION_ID and then ID (to decide if Debian or Ubuntu or ...)
            _, _ = self._session.execute_command('grep -i "^ID.*=debian"'
                                                     ' /etc/os-release')
            _, _ = self._session.execute_command('ls -ld /etc/apt')
        except CommandError as exc:
            lgr.debug("Did not detect Debian (or derivative): %s", exc)
            return

        packages, remaining_files = self.identify_packages_from_files(files)
        # TODO: add option to report distribution even if no packages/files
        # found association
        if not packages:
            return

        packages = self.get_details_for_packages(packages)

        # TODO: Depending on ID might be debian or ubuntu -- we might want to
        # absorb them all within DebianDistribution or have custom classes??
        # So far they seems to be largely similar
        dist = DebianDistribution(
            name="debian",
            version=debian_version,
            packages=packages,
            # TODO: helper to go from list -> dict based on the name, since must be unique
            apt_sources=list(self._apt_sources.values())
        )  # the one and only!
        dist.normalize()
        #   similar to DBs should take care about identifying/grouping etc
        #   of origins etc
        yield dist, remaining_files

    def _get_packagefields_for_files(self, files):
        # Call dpkg query in batches
        exec_gen = execute_command_batch(
            self._session, ['dpkg-query', '-S'], files,
            cmd_err_filter('no path found matching pattern'))
        # Parse and accumulate stat results in a dict
        file_to_package_dict = {}
        for (out, _, exc) in exec_gen:
            if exc:
                out = exc.stdout  # One file not found, so continue
            # Now go through the output and assign packages to files
            for outline in out.splitlines():
                # Parse package name (architecture) and path
                # TODO: Handle query of /bin/sh better
                outdict = self._parse_dpkgquery_line(outline)
                if not outdict:
                    lgr.debug("Skipping line %s", outline)
                    continue
                # Pull the found path from the dictionary
                found_name = outdict.pop('path')
                if not found_name:
                    raise ValueError(
                        "Record %s got no path defined... skipping"
                        % repr(outdict)
                    )
                # Associate the file to the package name (and architecture)
                pkg = outdict
                lgr.debug("Identified file %r to belong to package %s",
                          found_name, pkg)
                file_to_package_dict[found_name] = pkg
        return file_to_package_dict

    def _get_apt_source_name(self, src):
        # Create a unique name for the origin
        name_fmt = "apt_%s_%s_%s_%%d" % (src.origin or "", src.archive or "",
                                         src.component or "")
        name = utils.generate_unique_name(name_fmt, self._apt_source_names)
        # Remember the created name
        self._apt_source_names.add(name)
        # Create a named origin
        return name

    def get_details_for_packages(self, packages):
        if not packages:
            return []

        # Find apt sources if not defined
        if not self._all_apt_sources:
            self._find_all_sources()

        # Store the package details as dicts so that we can easily add to them
        pkg_dicts = [attr.asdict(pkg) for pkg in packages]

        # Use dpkg -s <pkg> to get arch and version
        self._get_pkgs_arch_and_version(pkg_dicts)

        # Use apt-cache show <pkg> to get details
        self._get_pkgs_details_from_apt_cache_show(pkg_dicts)

        # Now use "apt-cache policy pkg:arch" to get versions
        self._get_pkgs_versions_and_sources(pkg_dicts)

        # Get install date from the modify time of the dpkg info file
        self._get_pkgs_install_date(pkg_dicts)

        new_packages = []
        for p in pkg_dicts:
            new_pkg = DEBPackage(**p)
            new_packages.append(new_pkg)

        return new_packages

    def _create_package(self, name, architecture=None):

        # Store the details we currently know, we will populate the rest later
        return DEBPackage(
            name=name,
            architecture=architecture,
        )

    def _find_all_sources(self):
        # Use apt-cache policy to get all sources
        out, _ = self._session.execute_command(
            ['apt-cache', 'policy']
        )
        out = utils.to_unicode(out, "utf-8")

        src_info = parse_apt_cache_policy_source_info(out)
        for src_name in src_info:
            src_vals = src_info[src_name]
            date = self._get_date_from_release_file(
                src_vals.get("archive_uri"), src_vals.get("uri_suite"))
            self._all_apt_sources[src_name] = \
                APTSource(
                    name=src_name,
                    component=src_vals.get("component"),
                    codename=src_vals.get("codename"),
                    archive=src_vals.get("archive"),
                    architecture=src_vals.get("architecture"),
                    origin=src_vals.get("origin"),
                    label=src_vals.get("label"),
                    site=src_vals.get("site"),
                    date=date,
                    archive_uri=src_vals.get("archive_uri"))

    def _get_pkgs_arch_and_version(self, pkg_dicts):
        # Convert package names to name:arch format
        # Use "dpkg -s pkg" to get the installed version and arch
        # Note: "architecture" is in the dict, but may be null
        queries = [(p["name"] if not p["architecture"]
                    else "%(name)s:%(architecture)s" % p)
                   for p in pkg_dicts]
        # Call "dpkg -s" in batches
        exec_gen = execute_command_batch(self._session, ['dpkg', '-s'],
                                         queries)
        # Parse and accumulate "dpkg -s" results
        # dpkg -s uses the same output as apt-cache show pkg
        results = (parse_apt_cache_show_pkgs_output(out)
                   for (out, _, _) in exec_gen)
        # Combine sequence of lists
        results = itertools.chain.from_iterable(results)
        # Turn dpkg -s results into a lookup table by package name
        results = self.create_lookup_from_apt_cache_show(results)
        # Loop through each package and find the respective dpkg results
        for p in pkg_dicts:
            r = results.get(p["name"] if not p["architecture"]
                            else "%(name)s:%(architecture)s" % p)
            if not r:
                lgr.warning("Was unable to run dpkg -s for %s" %
                            p["name"])
                continue
            # Update the dictionary with found results
            p["architecture"] = r["architecture"]
            p["version"] = r["version"]

    @staticmethod
    def create_lookup_from_apt_cache_show(cmd_results):
        lookup_results = {}
        for r in cmd_results:
            lookup_results[r["package"]] = r
            lookup_results["%(package)s:%(architecture)s" % r] = r
        return lookup_results

    def _get_pkgs_details_from_apt_cache_show(self, pkg_dicts):
        # Convert package names to name:arch=version format
        queries = ["%(name)s:%(architecture)s=%(version)s" % p
                   for p in pkg_dicts]
        # Call "apt-cache show" in batches
        exec_gen = execute_command_batch(self._session, ['apt-cache', 'show'],
                                         queries)
        # Parse and accumulate "apt-cache show" results
        results = (parse_apt_cache_show_pkgs_output(out)
                   for (out, _, _) in exec_gen)
        # Combine sequence of lists
        results = itertools.chain.from_iterable(results)
        # Turn apt-cache show results into a lookup table by package name
        results = self.create_lookup_from_apt_cache_show(results)
        # Loop through each package and find the respective apt-cache results
        for p in pkg_dicts:
            r = results.get("%(name)s:%(architecture)s" % p)
            if not r:
                lgr.warning("Was unable to run apt-cache show for %s" %
                            p["name"])
                continue
            # Update the dictionary with found results (if present)
            for f in ("source_name", "source_version", "size", "md5",
                      "sha1", "sha256"):
                if f in r:
                    p[f] = r[f]

    def _get_pkgs_install_date(self, pkg_dicts):
        # Convert package names to dpkg list filenames
        queries = [self._pkg_name_to_dpkg_list_file(p["name"])
                   for p in pkg_dicts]
        # Call stat in batches
        exec_gen = execute_command_batch(
            self._session, ['stat', '-c', '%n: %Y'], queries,
            cmd_err_filter('No such file or directory'))
        # Parse and accumulate stat results in a dict
        results = {}
        for (out, _, exc) in exec_gen:
            if exc:
                out = exc.stdout  # One file not found, so continue
            # Parse the output and store by filename
            for outlines in out.splitlines():
                (fname, ftime) = outlines.split(": ")
                results[fname] = str(
                    pytz.utc.localize(
                        datetime.utcfromtimestamp(float(ftime))))

        # Now lookup the packages in the results
        for p in pkg_dicts:
            fname = self._pkg_name_to_dpkg_list_file(p["name"])
            if fname in results:
                p["install_date"] = results[fname]

    @staticmethod
    def _pkg_name_to_dpkg_list_file(name):
        query = "/var/lib/dpkg/info/" + name + ".list"
        return query

    def _get_pkgs_versions_and_sources(self, pkg_dicts):
        # Convert package names to name:arch format
        queries = ["%(name)s:%(architecture)s" % p for p in pkg_dicts]
        # Call apt-cache policy in batches
        exec_gen = execute_command_batch(self._session,
                                         ['apt-cache', 'policy'],
                                         queries)
        # Parse results into a single generator
        results = (parse_apt_cache_policy_pkgs_output(out)
                   for (out, _, _) in exec_gen)
        # Combine sequence of dicts
        results = join_sequence_of_dicts(results)
        # Loop through each package and find the respective apt-cache results
        for p in pkg_dicts:
            ver = results.get("%(name)s:%(architecture)s" % p)
            if not ver:
                ver = results.get(p["name"])
            if not ver:
                lgr.warning("Was unable to get version table for %s" %
                            p["name"])
                continue
            # Now construct the version table
            ver_dict = {}
            for v in ver.get("versions"):
                key = v["version"]
                ver_dict[key] = []
                for s in v.get("sources"):
                    s = s["source"]
                    # If we haven't named the source yet, name it
                    if s not in self._source_line_to_name_map:
                        # Make sure we can find the source
                        if s not in self._all_apt_sources:
                            lgr.warning("Cannot find source %s" % s)
                            continue
                        # Grab and name the source
                        source = self._all_apt_sources[s]
                        src_name = self._get_apt_source_name(source)
                        source.name = src_name
                        # Now add the source to our used sources
                        self._apt_sources[src_name] = source
                        # add the name for easy future lookup
                        self._source_line_to_name_map[s] = src_name
                    # Look up and add the short name for the source
                    ver_dict[key].append(
                        self._source_line_to_name_map[s])
            p["versions"] = ver_dict

    def _get_date_from_release_file(self, archive_uri, uri_suite):
        date = None
        for filename in get_apt_release_file_names(archive_uri, uri_suite):
            try:
                out = self._session.read(filename)
                spec = get_spec_from_release_file(out)
                date = str(pytz.utc.localize(
                    datetime.utcfromtimestamp(
                        mktime_tz(parsedate_tz(spec.date)))))
            except CommandError as _:
                # NOTE: We will be trying release files that end in
                # "Release" and "InRelease", so we expect to fail in opening
                # specific attempts.
                pass
        return date

    def _parse_dpkgquery_line(self, line):
        res = parse_dpkgquery_line(line)
        if res and res.pop("pkgs_rest"):
            if self._session.isdir(res["path"]):
                return None
            lgr.warning("dpkg-query line has multiple packages (%s)", line)
        return res



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Various supporting utilities for Debian(-based) distributions

"""

from __future__ import absolute_import

import logging
import re
import string

import attr

from niceman.utils import attrib

lgr = logging.getLogger('niceman.distributions.debian')

__docformat__ = 'restructuredtext'


@attr.s
class DebianReleaseSpec(object):
    """Encapsulates knowledge about used Debian release

    origin: Debian
    label: Debian
    suite: stable
    version: 8.5
    codename: jessie
    date: Sat, 04 Jun 2016 13:24:54 UTC
    components: main contrib non-free
    architectures: amd64
    """

    # Those could be different in overlays or other derived distributions
    origin = attrib(default=attr.NOTHING)   # Debian
    label = attrib(default=attr.NOTHING)    # Debian
    codename = attrib(default=attr.NOTHING)
    suite = attrib(default=attr.NOTHING)
    version = attrib(default=attr.NOTHING)
    date = attrib(default=attr.NOTHING)
    components = attrib(default=attr.NOTHING)
    architectures = attrib(default=attr.NOTHING)


def get_spec_from_release_file(content):
    """Provide specification object describing the component of the distribution
    """
    # RegExp to pull a single line "tag: value" pair from a deb822 file
    re_deb822_single_line_tag = re.compile("""
        ^(?P<tag>[a-zA-Z][^:]*):[\ ]+  # Tag - begins at start of line
        (?P<val>\S.*)$           # Value - after colon to the end of the line
    """, flags=re.VERBOSE + re.MULTILINE)

    # Split before PGP signature if present
    content = content.split("-----BEGIN PGP SIGNATURE-----")[0]

    # Parse the content for tags and values into a dictionary
    release = {
        match.group("tag"): match.group("val")
        for match in re_deb822_single_line_tag.finditer(content)
    }

    # TODO: redo with conversions of components and architectures in into lists
    # and date in machine-readable presentation
    return DebianReleaseSpec(**{
        a.name: release.get(a.name.title(), None)
        for a in attr.fields(DebianReleaseSpec)
        })


def parse_apt_cache_show_pkgs_output(output):
    package_info = []
    # Split into entries (one per package)
    entries = filter(bool, re.split('\n(?=Package:)', output,
                                    flags=re.MULTILINE))

    # RegExp to pull a single line "tag: value" pair from a deb822 file
    re_deb822_single_line_tag = re.compile("""
        ^(?P<tag>[a-zA-Z][^:]*):[\ ]+  # Tag - begins at start of line
        (?P<val>\S.*)$           # Value - after colon to the end of the line
    """, flags=re.VERBOSE + re.MULTILINE)
    # RegExp to split source into source and version
    re_source = re.compile("""
        ^(?P<source_name>[^ ]+)                # source name before any space
        ([^(]*\((?P<source_version>[^)]+)\))?  # source version in parentheses
    """, flags=re.VERBOSE)

    # For each package entry, collect single line tag/value pairs into a
    # dictionary
    for entry in entries:
        pkg = {
           match.group("tag").lower(): match.group("val")
           for match in re_deb822_single_line_tag.finditer(entry)
        }
        # Process the package if one was found
        if "package" in pkg:
            # Parse source line to get source version (if present)
            if "source" in pkg:
                for match in re_source.finditer(pkg["source"]):
                    pkg["source_name"] = match.group("source_name")
                    pkg["source_version"] = match.group("source_version")
            # Move md5sum to md5
            pkg["md5"] = pkg.pop("md5sum", None)
            # Append package entry
            package_info.append(pkg)
    return package_info


def parse_apt_cache_policy_pkgs_output(output):
    # findall wasn't greedy enough for some reason, so decided first to
    # split into entries (one per package)
    entries = filter(bool, re.split('\n(?=\S)', output, flags=re.MULTILINE))
    # now we need to parse/match each entry
    re_pkg = re.compile("""
        ^(?P<name>[^\s:]+):((?P<architecture>\S+):)?\s*\n  # package name
        \s+Installed:\s*(?P<installed>\S*)\s*\n    # Installed version
        \s+Candidate:\s*(?P<candidate>\S*)\s*\n    # Candidate version
        \s+Version\stable:[^\n]*
        (?P<version_table>(\n\s.*)+)
        """, flags=re.VERBOSE)

    re_versions = re.compile("""
        ^(\s{5}|\s(?P<installed>\*\*\*)\s)
        (?P<version>\S+)\s+
        (?P<priority>\S+)\n.*
        (?P<sources>(\n\s{8}.*)+)
    """, flags=re.VERBOSE + re.MULTILINE)
    re_source = re.compile("""
        ^\s{8}(?P<priority>\S+)\s+
        (?P<source>.*)$
    """, flags=re.VERBOSE + re.MULTILINE)
    pkgs = {}
    for entry in entries:
        match = re_pkg.match(entry.strip())
        if not match:
            lgr.warning("FAILED in %s " % entry)
            continue
        info = match.groupdict()
        pkgs[info.pop('name')] = info
        info['versions'] = []
        for version in re_versions.finditer(info.pop('version_table')):
            version_dict = version.groupdict()
            version_dict["sources"] = [
                source.groupdict()
                for source in re_source.finditer(version.group("sources"))
                ]
            info['versions'].append(version_dict)
        # process version_table
    return pkgs
    #for pkg_match in re_pkg.finditer(output):
    #    print type(pkg_match), pkg_match.groupdict()


def parse_apt_cache_policy_source_info(policy_output):
    source_info = {}
    re_section = re.compile("""
        ^(?P<header_line>\S.*)$[\r\n]*  # Header - non whitespace at the
                                        #          beginning of the line
        (?P<body>(^\ .*$[\r\n]*)*)      # Body - All subsequent lines that
                                        #        begin with a space
        """, flags=re.VERBOSE + re.MULTILINE)
    re_source = re.compile("""
        ^(\ (?P<priority>[0-9]+)\ +(?P<source>.*)$[\r\n]+
         (^(
            (\ \ +release\ +(?P<release_info>.*))|
            (\ \ +origin\ +(?P<origin_info>.*))|
            (\ \ .*)
           )$[\r\n]+
          )*
        )
        """, flags=re.VERBOSE + re.MULTILINE)
    re_rel_attrib = re.compile("""
        (?P<tag>[a-z])=    # A tag is a single letter followed by "="
        (?P<value>([^,]|(,(?![a-z]=)))*)
                           # The value follows the "=", and include any
                           # non commas, or commas not followed by another tag
        """, flags=re.VERBOSE)
    re_source_line = re.compile("""
        (?P<archive_uri>\S+)        # Archive URI up to the first " "
        (\ (?P<uri_suite>[^/]+))?   # The suite goes up to the first "/"
        """, flags=re.VERBOSE)
    # The release line has a terse tag=value format. This maps
    # the release tags to more meaningful values
    tag_map = {"c": "component",
               "n": "codename",
               "a": "archive",
               "b": "architecture",
               "o": "origin",
               "l": "label"}

    sections = re_section.finditer(policy_output)
    for section in sections:
        if section.group("header_line").startswith("Package files:"):
            sources = re_source.finditer(section.group("body"))
            for source in sources:
                info = source.groupdict()
                if info.get("source"):
                    src_detail = dict()
                    src_detail["site"] = info.get("origin_info")
                    match = re_source_line.search(info.get("source"))
                    if match:
                        match = match.groupdict()
                        src_detail["archive_uri"] = match["archive_uri"]
                        src_detail["uri_suite"] = match["uri_suite"]
                    else:
                        lgr.warning("Unexpected source line %s" %
                                    info.get("source"))
                    release_info = info.get("release_info")
                    if release_info:
                        attribs = re_rel_attrib.finditer(release_info)
                        for attrib in attribs:
                            if attrib.group("tag") in tag_map:
                                src_detail[tag_map[attrib.group("tag")]] = \
                                    attrib.group("value")
                    source_info[info.get("source")] = src_detail
    return source_info


def get_apt_release_file_names(url, url_suite):
    url = url.strip("/")              # Remove any trailing /
    url = url.replace("http://", "")  # Remove leading http://
    url = url.replace("file:/", "_")  # file:/ is converted to single _
    url = url.replace("/", "_")       # Any other / becomes _
    if url_suite:
        filename = url + "_dists_" + url_suite
    else:
        filename = url
    return ["/var/lib/apt/lists/" + filename + "_Release",
            "/var/lib/apt/lists/" + filename + "_InRelease"]


def parse_dpkgquery_line(line):
    result_re = re.compile(
        "(?P<name>[^,:]+)(:(?P<architecture>[^,:]+))?(?P<pkgs_rest>,.*)?:"
        " (?P<path>.*)$"
    )
    if line.startswith('diversion '):
        return None  # we are ignoring diversion details ATM  TODO

    res = result_re.match(line)
    if res:
        res = res.groupdict()
        if res['architecture'] is None:
            res.pop('architecture')
    return res



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Classes to manage compute resources."""

import attr
from importlib import import_module
import abc
from six import add_metaclass
from six.moves.configparser import NoSectionError

import yaml
from glob import glob
import os
import os.path as op

from ..dochelpers import exc_str
from ..support.exceptions import ResourceError
from ..support.exceptions import ResourceNotFoundError
from ..support.exceptions import ResourceAlreadyExistsError
from ..support.exceptions import MultipleResourceMatches
from ..support.exceptions import MissingConfigError


import logging
lgr = logging.getLogger('niceman.resource.base')


def get_required_fields(cls):
    """Return the mandatory fields for a resource class.
    """
    return {f.name for f in attr.fields(cls) if f.default is attr.NOTHING}


def get_resource_backends(cls):
    """Return name to documentation mapping of `cls`s backends.
    """
    return {b.name: b.metadata["doc"] for b in attr.fields(cls)
            if "doc" in b.metadata}


def backend_check_parameters(cls, keys):
    """Check whether any backend parameter keys are unknown.

    Parameters
    ----------
    cls : Resource object
    keys : iterable
        Backend parameter keys to check.

    Raises
    ------
    ResourceError on the first unknown key encountered.
    """
    required_params = get_required_fields(cls)
    for req_param in required_params:
        if req_param not in keys:
            raise ResourceError(
                "Missing required backend parameter: " + req_param)
    known = get_resource_backends(cls)
    for key in keys:
        if key not in known and key not in required_params:
            if known:
                import difflib

                suggestions = {s: known[s]
                               for s in difflib.get_close_matches(key, known)}
                if suggestions:
                    title = "Did you mean?"
                    params = suggestions
                else:
                    title = "Known backend parameters:"
                    params = known
                help_msg = "\n  {}\n{}\n".format(
                    title,
                    "\n".join(["    {} ({})".format(bname, bdoc)
                               for bname, bdoc in sorted(params.items())]))
                msg = "Bad --backend parameter '{}'{}".format(key, help_msg)
            else:
                msg = "Resource {} has no known parameters".format(cls)
            raise ResourceError(msg)


class ResourceManager(object):
    """Manager of available resources.

    Provides an API for finding existing resources or allocating new ones.
    Typically a NICEMAN process will have a single ResourceManager instance.
    """

    # The keys which are known to be secret and should not be exposed
    SECRET_KEYS = ('access_key_id', 'secret_access_key')

    def __init__(self, inventory_path=None):
        # Import here rather than the module-level to allow more flexibility in
        # overriding the configuration.
        from .. import cfg
        self.config_manager = cfg
        if inventory_path is None:
            self._inventory_path = self.config_manager.getpath(
                'general', 'inventory_file',
                op.join(cfg.dirs.user_config_dir, 'inventory.yml'))
        else:
            self._inventory_path = inventory_path

        self.inventory = self._get_inventory()

    def __iter__(self):
        for r in self.inventory:
            yield r

    @staticmethod
    def factory(config):
        """Factory method for creating the appropriate Container sub-class.

        Parameters
        ----------
        config : dict
            Configuration parameters for the resource.

        Returns
        -------
        Resource sub-class instance.
        """
        if 'type' not in config:
            raise MissingConfigError("Resource 'type' parameter missing for resource.")

        type_ = config['type']
        module_name = '_'.join(type_.split('-'))
        class_name = ''.join([token.capitalize() for token in type_.split('-')])
        try:
            module = import_module('niceman.resource.{}'.format(module_name))
        except Exception as exc:
            # Typically it should be an ImportError, but let's catch and recast
            # anything just in case.
            try:
                msg = exc_str(exc)
                known = ResourceManager._discover_types()
                if module_name not in known:
                    msg += ". Known ones are: {}".format(", ".join(known))
            except Exception as exc2:
                msg += ".  Failed to discover resource types: " + exc_str(exc2)
            raise ResourceError(
                "Failed to import resource: {}".format(msg)
            )
        cls = getattr(module, class_name)
        try:
            instance = cls(**config)
        except TypeError:
            backend_check_parameters(cls, config)
            # The check didn't raise an exception, so this wasn't related to an
            # unknown backend parameter.
            raise
        return instance

    # TODO: Following methods might better be in their own class
    @staticmethod
    def _discover_types():
        """Discover resource types by inspecting the resource directory files.

        Returns
        -------
        string list
            List of resource identifiers extracted from file names.
        """
        l = []
        for f in glob(op.join(op.dirname(__file__), '*.py')):
            f_ = op.basename(f)
            if f_ in ('base.py',) or f_.startswith('_'):
                continue
            l.append(f_[:-3])
        return sorted(l)

    def _find_resources(self, resref, resref_type):
        def from_name(x):
            return [(name, config) for name, config in self.inventory.items()
                    if x == name]

        def from_id(x):
            return [(name, config) for name, config in self.inventory.items()
                    if x == config.get("id")]

        results_name = None
        results_id = None
        if resref_type == "auto":
            results_name = from_name(resref)
            results_id = from_id(resref)
        elif resref_type == "name":
            results_name = from_name(resref)
        elif resref_type == "id":
            results_id = from_id(resref)
        return results_name, results_id

    def _get_resource_config(self, resref, resref_type="auto"):
        results_name, results_id = self._find_resources(resref, resref_type)
        if results_name and results_id:
            raise MultipleResourceMatches(
                "{} is ambiguous. "
                "Explicitly specify whether it is a name or id".format(resref))
        elif not (results_name or results_id):
            raise ResourceNotFoundError(
                "Resource matching {} not found".format(resref))
        elif results_id and len(results_id) > 1:
            raise MultipleResourceMatches(
                "ID {} matches multiple resources. "
                "Try specifying with the name instead".format(resref))

        name, inventory_config = (results_name or results_id)[0]
        type_ = inventory_config['type']
        try:
            config = dict(self.config_manager.items(type_.split('-')[0]))
        except NoSectionError:
            config = {}
        config.update(inventory_config)
        return config

    def get_resource(self, resref, resref_type="auto"):
        """Return the resource instance for `resref`.

        Parameters
        ----------
        resref : str
            A name or ID that uniquely maps to a resource.
        resref_type : {'auto', 'name', 'id'}, optional
            The default behavior is to infer whether `resref` refers to a
            resource by name or full ID.  If the reference is ambiguous, this
            fails, in which case 'name' or 'id' can be used to disambiguate
            `resref`.

        Returns
        -------
        A Resource object
        """
        return self.factory(self._get_resource_config(resref, resref_type))

    def _get_inventory(self):
        """Return a dict with the config information for all resources.

        Parameters
        ----------
        inventory_path : string
            Path to the inventory file which is declared in the niceman.cfg file.

        Returns
        -------
        inventory : dict
            Hash whose key is resource name and value is the config settings for
            the resource.
        """
        inventory_path = self._inventory_path
        if not inventory_path:
            raise MissingConfigError(
                "No resource inventory path is known to %s" % self
            )

        if not op.isfile(inventory_path):
            inventory = {}
        else:
            with open(inventory_path, 'r') as fp:
                inventory = yaml.safe_load(fp)

        return inventory

    def _save(self):
        """Save the resource inventory.
        """
        # Operate on a copy so there is no side-effect of modifying original
        # inventory.
        #
        # The attribute may not exist yet because _get_inventory calls _save.
        inventory = self.inventory.copy() if hasattr(self, "inventory") else {}

        for key in list(inventory):  # go through a copy of all keys since we modify

            # A resource without an ID has been deleted.
            inventory_item = inventory[key]
            if 'id' in inventory_item and not inventory_item['id']:
                del inventory[key]

            # Remove AWS credentials
            # TODO(yoh): split away handling of credentials.  Resource should
            # probably just provide some kind of an id for a credential which
            # should be stored in a safe credentials storage
            for secret_key in ResourceManager.SECRET_KEYS:
                if secret_key in inventory_item:
                    del inventory_item[secret_key]

        if not op.exists(op.dirname(self._inventory_path)):
            os.makedirs(op.dirname(self._inventory_path))

        with open(self._inventory_path, 'w') as fp:
            yaml.safe_dump(inventory, fp, default_flow_style=False)

    def create(self, name, resource_type, backend_params=None):
        results_name, results_id = self._find_resources(name, "auto")
        if results_name or results_id:
            raise ResourceAlreadyExistsError(
                "Resource with {} {} already exists"
                .format("name" if results_name else "ID", name))

        config = {'name': name, 'type': resource_type}
        if backend_params:
            config.update(backend_params)
        resource = self.factory(config)
        resource.connect()
        resource_attrs = resource.create()
        config.update(resource_attrs)
        self.inventory[name] = config
        self._save()

    def delete(self, resource):
        """Delete `resource` from the inventory.

        Parameters
        ----------
        resource : Resource object
        """
        resource.delete()
        del self.inventory[resource.name]
        self._save()

    def start(self, resource):
        """Start a `resource` in the inventory.

        Parameters
        ----------
        resource : Resource object
        """
        try:
            resource.start()
        except NotImplementedError:
            lgr.info("This resource type does not support the 'start' feature")
            return

        self.inventory[resource.name]['status'] = "running"
        self._save()
        lgr.info("Started the environment %s (%s)", resource.name, resource.id)

    def stop(self, resource):
        """Stop but do not delete a `resource` in the inventory.

        Parameters
        ----------
        resource : Resource object
        """
        try:
            resource.stop()
        except NotImplementedError:
            lgr.info("This resource type does not support the 'stop' feature")
            return

        self.inventory[resource.name]['status'] = "stopped"
        self._save()
        lgr.info("Stopped the environment %s", resource.name)


@add_metaclass(abc.ABCMeta)
class Resource(object):
    """Base class for creating and managing compute resources.
    """

    def __repr__(self):
        return 'Resource({})'.format(self.name)

    def add_command(self, command, env=None):
        """Add a command to the command buffer so that all commands can be
        run at once in a batch submit to the environment.

        Parameters
        ----------
        command : string or list
            Command string or list of command string tokens.

        env : dict
            Additional (or replacement) environment variables which are applied
            only to the current call
        """
        if not hasattr(self, '_command_buffer'):
            self._command_buffer = []  # Each element is a dictionary in the
                                       # form {command=[], env={}}
        self._command_buffer.append({'command': command, 'env': env})

    def execute_command_buffer(self, session=None):
        """Send all the commands in the command buffer to the environment for
        execution.

        Parameters
        ----------
        session : Sesson object, optional
            Session object reflects the resource type. (the default is None,
            which will cause the Session object to be retrieved from the
            Resource object.)
        """
        if not session:
            session = self.get_session(pty=False)
        for command in self._command_buffer:
            lgr.debug("Running command '%s'", command['command'])
            session.execute_command(command['command'], env=command['env'])

    def set_envvar(self, var, value):
        """Save an environment variable for inclusion in the environment

        Parameters
        ----------
        var : string
            Env variable name
        value : string
            Env variable value
        """
        # TODO: This wouldn't work correctly since pretty much each command
        # then should have its desired env recorded since set_envvar
        # could be interleaved with add_command calls
        if not hasattr(self, '_env'):
            self._env = {}

        self._env[var] = value

    def get_updated_env(self, custom_env):
        """Returns an env dictionary with additional or replaced values.

        Parameters
        ----------
        custom_env : dict
            Environment variables to merge into the existing list of declared
            environment variables stored in self._env

        Returns
        -------
        dict
            Environment variables merged with additional custom variables.
        """
        if hasattr(self, '_env'):
            merged_env = self._env.copy()
            if custom_env:
                merged_env.update(custom_env)
            return merged_env

        return custom_env

    @classmethod
    def _generate_id(cls):
        """Utility class method to generate a UUID.

        Returns
        -------
        string
            Newly created UUID
        """
        import uuid
        # just a random uuid for now, TODO: think if we somehow could
        # fingerprint it so to later be able to decide if it is 'ours'? ;)
        return str(uuid.uuid1())

    @abc.abstractmethod
    def get_session(self, pty=False, shared=None):
        """Returns the Session object for this resource.

        Parameters
        ----------
        pty : bool, optional
            Terminal session (the default is False)
        shared : string, optional
            Shared session identifier (the default is None)
        """
        return



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper utility to run commands in an environment
"""

__docformat__ = 'restructuredtext'

import os
import os.path as op
import sys
import time
import uuid

from .base import Interface
from ..support.exceptions import CommandError
from ..support.exceptions import MissingExternalDependency
import niceman.interface.base  # Needed for test patching
from ..support.param import Parameter
from ..support.constraints import EnsureStr
from ..support.external_versions import external_versions
from ..resource import get_manager
from ..resource.session import Session
from .common_opts import trace_opt
from .common_opts import resref_opt
from .common_opts import resref_type_opt

from logging import getLogger
lgr = getLogger('niceman.api.execute')


class CommandAdapter(object):
    """Interface to kinds of commands `exec` can execute.

    Each subclass should define an `execute` method and optionally define
    `pre_command` or `post_command` to perform pre- or post-command operations.
    The caller should execute the command by calling the instance object
    itself.

    Parameters
    ----------
    resource : Resource object
    command : str
        The program to run.
    cmd_args : list of str
        Arguments to pass to `command`.
    """

    def __init__(self, resource, command, cmd_args):
        self.resource = resource
        self.session = resource.get_session()
        self.command = command
        self.cmd_args = cmd_args

    def pre_command(self):
        pass

    def execute(self):
        """Run the full command defined by `command` and `cmd_args`.

        Returns
        -------
        A tuple, (stdout, stderr).
        """
        raise NotImplementedError

    def post_command(self):
        pass

    def __call__(self):
        self.pre_command()
        try:
            out, err = self.execute()
        finally:
            self.post_command()
        return out, err


class PlainCommand(CommandAdapter):

    def execute(self):
        return self.session.execute_command([self.command] + self.cmd_args)


class InternalCommand(CommandAdapter):

    def execute(self):
        self.session.niceman_exec(self.command, self.cmd_args)
        return None, None  # Return same form as Session.execute_command.


class TracedCommand(CommandAdapter):
    """A tracer-wrapped command.

    In the pre-command step, a tracer (extracted from ReproZip) is downloaded
    to the local machine if not already present.  This tracer is transferred to
    the resource on which the command will be executed.  This allows tracing on
    a remote resource that does not have ReproZip installed.

    After the command is executed under the tracer, the post-command step
    downloads the trace artifacts locally, uses ReproZip to write a
    configuration file from these artifacts, and then calls `niceman retrace`
    on the result..
    """

    def __init__(self, resource, command, cmd_args,
                 remote_dir=None, local_dir=None):
        super(TracedCommand, self).__init__(resource, command, cmd_args)

        if not external_versions["reprozip"]:
            raise MissingExternalDependency("Using --trace requires ReproZip, "
                                            "a Linux-specific dependency")

        self.tracer_md5sum = "d8561c1bc528592b21c0e28d6f32c0a4"
        # adding two random characters to avoid collisions etc
        # The id for the execution so we could pick up all the log and trace
        # files for local storage
        self.exec_id = "{}-{}".format(
            time.strftime("%Y%m%d%H%M%S"),
            str(uuid.uuid4())[:2])

        # Local session variables
        local_cache_dir = local_dir or op.expanduser('~/.cache/niceman')
        self.local_tracer_dir = op.join(local_cache_dir,
                                        "tracers",
                                        self.tracer_md5sum)
        self.local_trace_dir = op.join(local_cache_dir, 'traces', self.exec_id)
        self.local_tracer_gz = op.join(self.local_tracer_dir,
                                       "niceman_trace.gz")

        # Remote session variables
        self.remote_dir = remote_dir
        self.remote_trace_dir = None
        self.remote_tracer = None

    def _prepare_local(self):
        if not op.exists(self.local_tracer_gz):
            import hashlib
            import requests

            if not op.exists(self.local_tracer_dir):
                os.makedirs(self.local_tracer_dir)

            lgr.info("Downloading tracer...")
            resp = requests.get("https://github.com/ReproNim/reprozip/blob"
                                "/0497b229575c67219c5925360b6e63bf8d4d5eb9"
                                "/reprozip/native/rztracer.gz?raw=true",
                                allow_redirects=True)

            with open(self.local_tracer_gz, "wb") as stream:
                if self.tracer_md5sum != hashlib.md5(resp.content).hexdigest():
                    raise RuntimeError("md5sum for downloaded tracer "
                                       "does not match the expected one")
                stream.write(resp.content)
            lgr.info("Tracer downloaded to %s", self.local_tracer_gz)

    def _prepare_remote(self):
        # Establish a "management" session
        mng_ses = self.resource.get_session(pty=False)
        remote_env_full = mng_ses.query_envvars()
        root = self.remote_dir or '{HOME}/.cache'.format(**remote_env_full)
        remote_niceman_dir = '{}/niceman'.format(root)

        remote_traces_dir = op.join(remote_niceman_dir, 'traces')
        mng_ses.mkdir(remote_traces_dir, parents=True)
        self.remote_trace_dir = op.join(remote_traces_dir, self.exec_id)
        mng_ses.mkdir(self.remote_trace_dir, parents=True)

        remote_tracer_dir = op.join(remote_niceman_dir,
                                    "tracers",
                                    self.tracer_md5sum)
        # TODO: augment "entry point" somehow in a generic way?
        #    For interactive sessions with bash, we could overload ~/.bashrc
        #    to do our wrapping of actual call to bashrc under the "tracer"
        self.remote_tracer = op.join(remote_tracer_dir, "niceman_trace")

        if not self.session.exists(self.remote_tracer):
            remote_tracer_gz = self.remote_tracer + ".gz"
            # The gz file might already exist (e.g., a localshell session).
            if not self.session.exists(remote_tracer_gz):
                self.session.put(self.local_tracer_gz, remote_tracer_gz)
            self.session.execute_command(
                ["gunzip", "--keep", remote_tracer_gz])
            self.session.chmod(self.remote_tracer, "755")
        # TODO: might want to add also a "marker" so within the trace
        #       we could avoid retracing session establishing bits themselves

    def pre_command(self):
        self._prepare_local()
        self._prepare_remote()

    def execute(self):
        cmd_prefix = [
            self.remote_tracer,
            "--logfile", op.join(self.remote_trace_dir, "tracer.log"),
            "--dbfile", op.join(self.remote_trace_dir, "trace.sqlite3"),
            "--"
        ]
        return self.session.execute_command(
            cmd_prefix + [self.command] + self.cmd_args)

    def post_command(self):
        # Copy all the tracing artifacts here if not present already (e.g.
        # if session was a local shell)
        if not op.exists(self.local_trace_dir):
            for fname in ["tracer.log", "trace.sqlite3"]:
                self.session.get(op.join(self.remote_trace_dir, fname),
                                 op.join(self.local_trace_dir, fname))
            lgr.info(
                "Copied tracing artifacts under %s", self.local_trace_dir)
        else:
            lgr.debug(
                "Not copying %s from remote session "
                "since already exists locally",
                self.local_trace_dir)

        from reprozip.tracer.trace import write_configuration
        from rpaths import Path

        # we rely on hardcoded paths in reprozip
        write_configuration(
            directory=Path(self.local_trace_dir),
            sort_packages=False,
            find_inputs_outputs=True)

        from niceman.api import retrace
        niceman_spec_path = op.join(self.local_trace_dir, "niceman.yml")
        retrace(
            spec=op.join(self.local_trace_dir, "config.yml"),
            output_file=niceman_spec_path,
            resref=self.session
        )
        lgr.info("NICEMAN trace %s", niceman_spec_path)


# Exists for ease of testing.
CMD_CLASSES = {"plain": PlainCommand,
               "internal": InternalCommand,
               "trace": TracedCommand}


class Execute(Interface):
    """Execute a command in a computation environment

    Examples
    --------

      $ niceman execute mkdir /home/blah/data

    """

    _params_ = dict(
        command=Parameter(
            doc="name of the command to run",
            metavar='COMMAND',
            constraints=EnsureStr(),
        ),
        args=Parameter(
            doc="list of positional and keyword args to pass to the command",
            metavar='ARGS',
            nargs="*",
            constraints=EnsureStr(),
        ),
        resref=resref_opt,
        # XXX reenable when we support working with multiple instances at once
        # resource_type=Parameter(
        #     args=("-t", "--resource-type"),
        #     doc="""Resource type to work on""",
        #     constraints=EnsureStr(),
        # ),
        resref_type=resref_type_opt,
        # TODO: should be moved into generic API
        internal=Parameter(
            args=("--internal",),
            action="store_true",
            doc="Instead of running a generic/any command, execute the internal"
                " NICEMAN command available within sessions.  Known are: %s"
                % ', '.join(Session.INTERNAL_COMMANDS)
        ),
        trace=trace_opt,
    )

    @staticmethod
    def __call__(command, args, resref=None, resref_type="auto",
                 internal=False, trace=False):
        from niceman.ui import ui

        if internal and trace:
            raise NotImplementedError("No --trace for --internal commands")

        if not resref:
            resref = ui.question(
                "Enter a resource name or ID",
                error_message="Missing resource name or ID"
            )

        env_resource = get_manager().get_resource(resref, resref_type)
        env_resource.connect()

        if internal:
            cls_key = "internal"
        elif trace:
            cls_key = "trace"
        else:
            cls_key = "plain"
        cmd = CMD_CLASSES[cls_key](env_resource, command, args)

        try:
            error = None
            out, err = cmd()
        except CommandError as exc:
            error = exc
            out, err = exc.stdout, exc.stderr

        lgr.info("Executed the %s command in the environment %s", command,
                 env_resource.name)

        if out:
            sys.stdout.write(out)
        if err:
            sys.stderr.write(err)
        if error:
            lgr.error(
                "Command %s failed to run in %s: %s",
                command, env_resource.name, error.msg
            )
            raise SystemExit(error.code)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper utility to list available environments
"""

__docformat__ = 'restructuredtext'

from six.moves.configparser import NoSectionError

from .base import Interface
import niceman.interface.base # Needed for test patching
from ..support.param import Parameter
from ..support.constraints import EnsureStr, EnsureNone
from  ..resource import get_manager
from ..ui import ui
from ..support.exceptions import ResourceError
from ..support.exceptions import ResourceNotFoundError
from ..dochelpers import exc_str

from logging import getLogger
lgr = getLogger('niceman.api.ls')


class Ls(Interface):
    """List known computation resources, images and environments

    Examples
    --------

      $ niceman ls
    """

    _params_ = dict(
        verbose=Parameter(
            args=("-v", "--verbose"),
            action="store_true",
            #constraints=EnsureBool() | EnsureNone(),
            doc="provide more verbose listing",
        ),
        refresh=Parameter(
            args=("--refresh",),
            action="store_true",
            doc="Refresh the status of the resources listed",
            # metavar='CONFIG',
            # constraints=EnsureStr(),
        ),
    )

    @staticmethod
    def __call__(verbose=False, refresh=False):
        id_length = 19  # todo: make it possible to output them long
        template = '{:<20} {:<20} {:<%(id_length)s} {:<10}' % locals()
        ui.message(template.format('RESOURCE NAME', 'TYPE', 'ID', 'STATUS'))
        ui.message(template.format('-------------', '----', '--', '------'))

        manager = get_manager()
        for name in sorted(manager):
            if name.startswith('_'):
                continue

            # if refresh:
            try:
                resource = manager.get_resource(name, resref_type="name")
            except ResourceNotFoundError:
                lgr.warning("Manager did not return a resource for %r", name)
                continue

            try:
                if refresh:
                    resource.connect()
                if not resource.id:
                    # continue  # A missing ID indicates a deleted resource.
                    resource.id = 'DELETED'
                    resource.status = 'N/A'
                report_status = resource.status
            except Exception as exc:
                lgr.error("%s resource query error: %s", name, exc_str(exc))
                report_status = "N/A (QUERY-ERROR)"
                for f in 'id', 'status':
                    if not getattr(resource, f):
                        setattr(resource, f, "?")
            msgargs = (
                name,
                resource.type,
                resource.id[:id_length] if resource.id else '',
                report_status,
            )
            ui.message(template.format(*msgargs))
            lgr.debug('list result: {}, {}, {}, {}'.format(*msgargs))

        # if not refresh:
        #     ui.message('(Use --refresh option to view current status.)')
        #
        # if refresh:
        #     niceman.interface.base.set_resource_inventory(inventory)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Python NICEMAN API exposing user-oriented commands (also available via CLI)"""


def _generate_func_api():
    """Auto detect all available interfaces and generate a function-based
       API from them
    """
    from importlib import import_module
    from .interface.base import update_docstring_with_parameters
    from .interface.base import get_interface_groups
    from .interface.base import get_api_name
    from .interface.base import alter_interface_docs_for_api

    for grp_name, grp_descr, interfaces in get_interface_groups():
        for intfspec in interfaces:
            # turn the interface spec into an instance
            mod = import_module(intfspec[0], package='niceman')
            intf = getattr(mod, intfspec[1])
            spec = getattr(intf, '_params_', dict())

            # FIXME no longer using an interface class instance
            # convert the parameter SPEC into a docstring for the function
            update_docstring_with_parameters(
                intf.__call__, spec,
                prefix=alter_interface_docs_for_api(
                    intf.__doc__),
                suffix=alter_interface_docs_for_api(
                    intf.__call__.__doc__)
            )
            globals()[get_api_name(intfspec)] = intf.__call__

# Invoke above helpers
_generate_func_api()

# Be nice and clean up the namespace properly
del _generate_func_api



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Plugin support for provenance YAML files following NICEMAN spec.

"""
from __future__ import absolute_import

import collections
import datetime
import logging
from collections import OrderedDict

import attr
import yaml

import niceman
from niceman.distributions.base import Factory
from niceman.distributions.base import SpecObject
from niceman.utils import instantiate_attr_object
from .base import Provenance
from .utils import write_config
from .. import utils
from ..distributions import Distribution
from ..dochelpers import exc_str

lgr = logging.getLogger('niceman.formats.niceman')

__version__ = '0.0.1'


class NicemanProvenance(Provenance):
    """
    Parser for NICEMAN Spec (YAML specification)
    """

    @classmethod
    def _load(cls, source):
        """
        Load and store the raw spec file.

        Parameters
        ----------
        source : string
            File path or URL or actual spec if contains new line in it
        """
        # TODO: should we somehow provide a custom loader in case of
        # dictionaries to make them ordered dict?  initial use case
        # is listing of distributions. although it is not clear yet
        # either order should matter.  Now in some places then internally
        # sorting alphabetically for consistency
        if '\n' in source:
            return yaml.load(source)

        with open(source, 'r') as stream:
            try:
                return yaml.load(stream)
            except yaml.YAMLError as exc:
                lgr.error("Failed to load %s: %s", source, exc_str(exc))
                raise  # TODO -- we might want a dedicated exception here

    # def get_operating_system(self):
    #     """
    #     Retrieve the operating system information.
    #
    #     Returns
    #     -------
    #     Dictionary containing name and version of the OS.
    #         os['name']
    #         os['version']
    #     """
    #     raise NotImplementedError()
    #     return self._src['distribution']

    def get_distributions(self):
        """
        Retrieve the information for all the distributions recorded in the
        file.

        Parameters
        ----------
        in_value: list or dict or object
          Value (in native
        Returns
        -------
        list
            List of Distribution sub-class objects.
        """
        from ..distributions.base import TypedList
        return self._load_spec(self._src["distributions"], TypedList(Distribution))

    def _load_spec(self, in_value, factory_class=None):
        """
        Parameters
        ----------
        in_value: list or dict or object
          Value (in native
        factory_class: object, optional
          If not specified, we expect to receive a single item and deduce its
          type from the `name` attribute

        Returns
        -------
        SpecObject or list of SpecObject(s)

        """
        dist_objects = []

        """
        TODO: For "compressed" presentation we need to specify somewhere a handler
        in case of e.g. 
        - packages:
           - "full" is a list of dicts(attr structures)
           - a list with strings (e.g. - name[>=version]) where, depending on type
             of the package ('name' for DEBPackage, "url" for VCS), some name serves
             an identified
           - so it is a matter of the datatype of the item (string vs fully blown dict)
        - distributions:
           - "full" is a list of dict(attr structures) per each type of distribution
           - a dict with keys being a "name" of each entry
           
        But that is later -- for now just reading as is, allowing to map those specs
        into structs
        """

        """ATM we will assume that we are getting the "cut" at the level of the environment
        but it could be that the spec is on top of it
        """

        # TODO: may be we need to check if it has a type
        def is_TypedList(cls):
            # since _CountingAttr is protected we do ducktyping
            return hasattr(cls, 'metadata') \
                   and cls.metadata.get('type') \
                   and hasattr(cls, "_default") \
                   and issubclass(cls._default.factory, list)

        from attr import Attribute

        if is_TypedList(factory_class) or isinstance(factory_class, Attribute):
            if isinstance(in_value, dict):
                # normalize compressed presentation into full (sugaring)
                in_value = [
                    dict(name=n, **(fields or {}))
                    for n, fields in sorted(in_value.items())
                ]

            assert isinstance(in_value, list)
            # We have a factory which is a list of things
            return [
                self._load_spec(item, factory_class.metadata['type'])
                for item in in_value
            ]

        # A single item case
        assert not isinstance(in_value, (list, tuple))

        if hasattr(factory_class, "factory"):
            # "our" factory which we might have defined in our SpecObject classes
            subclass = in_value['name'].strip('-0123456789')
            # Uses our factory decided by the 'name'
            # So it is pretty much some kind of a helper factory
            #   get_instance_by_name('niceman.distributions', in_value['name'])
            # and then populate it.  Could become part of the model spec
            # describing that. ATM it is just a FactoryListOf(Distribution)
            # but we want to say that it is not just any Distribution
            # but the one decided from the 'name' and class for which found
            # among available in a module.
            # We have pretty much the same "factory" construct for Resources
            # ATM.
            # RF: make it generic!
            spec_class = factory_class.factory(subclass)
        else:
            spec_class = factory_class

        # arguments to instantiate the class
        spec_args = []
        spec_kwargs = dict()  # name=in_value['name'])

        # process fields known to the attr
        spec_attrs = spec_class.__attrs_attrs__  # as is -- list of them
        spec_in = in_value.copy()  # shallow copy so we could pop

        # now we need to see what fields are present in the spec,
        # and prepare them to be passed into its constructor
        for spec_attr in spec_attrs:
            name = spec_attr.name
            if name not in spec_in:
                if spec_attr.default is attr.NOTHING:
                    # positional argument -- must be known
                    raise ValueError(
                        "%s requires %r field, but was provided only with "
                        "following fields: %s"
                        % (spec_class.__name__, name, ', '.join(spec_in.keys()))
                    )
                else:
                    # skipping because we have no value and there is a default
                    continue

            # popping so later we could verify that we handled all keys in the
            # spec_in
            value_in = spec_in.pop(name)

            # now we need a "factory" for each of those records
            # And those could be specific to their type(s) when "compressed"
            # but in general we should be able to use the same logic,
            # just need to know whom to call
            if isinstance(spec_attr.default, Factory):
                item_type = spec_attr.metadata.get('type')
                if item_type:
                    # Recurse this whole shebang
                    value_out = self._load_spec(value_in, spec_attr)
                else:
                    # ATM only files could have a list of untyped entries
                    assert spec_attr.name in ('files', 'remotes')
                    value_out = instantiate_attr_object(spec_attr.default.factory, value_in)
            else:
                value_out = value_in

            # Positional arguments to attr is the ones without a default
            if spec_attr.default is attr.NOTHING:
                spec_args.append(value_out)  # positional arg
            else:
                spec_kwargs[spec_attr.name] = value_out  # keyword arg

        if spec_in:
            raise ValueError(
                "Following input fields were not processed since were not "
                "known to %s: %s"
                % (spec_class.__name__, ', '.join(spec_in.keys()))
            )

        return spec_class(*spec_args, **spec_kwargs)

    def get_files(self, limit='all'):
        files = self._src.get('files', [])
        if limit in ('all', 'packaged'):
            files.append('TODO')
            # TODO: we would need to get_distributions first then to traverse
            # all the packages etc...
        return files

    # TODO: RF
    #   config must be gone and taken from self
    @classmethod
    def write(cls, output, spec):
        """Writes an environment config to a stream
    
        Parameters
        ----------
        output
            Output Stream
        spec : dict
            Spec (environment) configuration (input).
            ??? Might want to code it in a generic fashion so spec 
            might be at a different level than environment may be.
            E.g. something which would be above and contain environment(s), 
            runs, etc 
    
        """

        # Allow yaml to handle OrderedDict
        # From http://stackoverflow.com/questions/31605131
        if collections.OrderedDict not in yaml.SafeDumper.yaml_representers:
            yaml.SafeDumper.add_representer(
                collections.OrderedDict,
                lambda self, data:
                self.represent_mapping('tag:yaml.org,2002:map', data.items()))

        utils.safe_write(
            output,
            ("# NICEMAN Environment Configuration File\n"
             "# This file was created by NICEMAN {0} on {1}\n").format(
                niceman.__version__, datetime.datetime.now()))

        #c = "\n# Runs: Commands and related environment variables\n\n"
        #write_config_key(output, envconfig, "runs", c)

        out = OrderedDict()
        out['version'] = __version__
        out.update(spec_to_dict(spec))
        write_config(output, out)
        return out


# TODO: RF into SpecObject._as_dict()
def spec_to_dict(spec):

    out = OrderedDict()
    spec_attrs = spec.__attrs_attrs__  # as is -- list of them
    for attr in spec_attrs:
        value_in = getattr(spec, attr.name, None)
        if not value_in:
            continue
        if isinstance(value_in, Factory):
            # wasn't set, thus "default", thus
            continue
        elif isinstance(value_in, list):
            # might be specs too
            value_out = value_in.__class__(
                spec_to_dict(v) if isinstance(v, SpecObject) else v
                for v in value_in
            )
        elif isinstance(value_in, SpecObject):
            value_out = spec_to_dict(value_in)
        else:
            value_out = value_in
        if value_out in (tuple(), [], {}, None):
            continue  # do not bother saving empty ones

        out[attr.name] = value_out
    return out

"""
        envconfig = dict(spec)  # Shallow copy for destruction

        c = "APT sources"
        write_config_key(output, envconfig, "apt_sources", c)

        c = "Packages"
        write_config_key(output, envconfig, "packages", c)

        c = "Non-Packaged Files"
        write_config_key(output, envconfig, "other_files", c)

        if envconfig:
            utils.safe_write(output, "\n# Other ReproZip keys (not used by NICEMAN) \n\n")
            utils.safe_write(output, yaml.safe_dump(envconfig,
                                                    encoding="utf-8",
                                                    allow_unicode=True))
"""





# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import os.path as op
import pytest

from niceman.resource.base import ResourceManager
from niceman.resource.base import backend_check_parameters
from niceman.resource.shell import Shell
from niceman.resource.docker_container import DockerContainer
from niceman.support.exceptions import MissingConfigError
from niceman.support.exceptions import MultipleResourceMatches
from niceman.support.exceptions import ResourceAlreadyExistsError
from niceman.support.exceptions import ResourceError


def test_resource_manager_factory_missing_type():
    with pytest.raises(MissingConfigError):
        ResourceManager.factory({})


def test_resource_manager_factory_unkown():
    with pytest.raises(ResourceError):
        ResourceManager.factory({"type": "not really a type"})


def test_backend_check_parameters_no_known():
    with pytest.raises(ResourceError) as exc:
        backend_check_parameters(Shell,
                                 {"name": "name",
                                  "unknown_key": "value"})
    assert "no known parameters" in str(exc.value)


def test_backend_check_parameters_nowhere_close():
    with pytest.raises(ResourceError) as exc:
        backend_check_parameters(DockerContainer,
                                 {"name": "name",
                                  "unknown_key": "value"})
    assert "Known backend parameters" in str(exc.value)


def test_backend_check_parameters_close_match():
    with pytest.raises(ResourceError) as exc:
        backend_check_parameters(DockerContainer,
                                 {"name": "name",
                                  "imagee": "value"})
    assert "Did you mean?" in str(exc.value)


def test_backend_check_parameters_missing_required():
    with pytest.raises(ResourceError) as exc:
        backend_check_parameters(DockerContainer, {"imagee": "value"})
    assert "Missing required" in str(exc.value)


def test_resource_manager_empty_init(tmpdir):
    inventory = op.join(str(tmpdir), "inventory.yml")
    manager = ResourceManager(inventory)
    assert not manager.inventory


def test_resource_manager_inventory_undefined():
    with pytest.raises(MissingConfigError):
        ResourceManager("")


def test_resource_manager_save(tmpdir):
    inventory = op.join(str(tmpdir), "subdir", "inventory.yml")
    manager = ResourceManager(inventory)
    manager.inventory = {
        "plain": {"name": "plain",
                  "type": "foo-type",
                  "id": "foo_id"},
        "with-secret": {"name": "with-secret",
                        "type": "bar-type",
                        "secret_access_key": "SECRET",
                        "id": "bar_id"},
        "null-id": {"name": "null-id",
                    "id": None,
                    "type": "baz-type"}}
    manager._save()
    assert op.exists(inventory)
    with open(inventory) as fh:
        content = fh.read()
    assert "plain" in content
    assert "with-secret" in content
    assert "SECRET" not in content
    assert "null-id" not in content

    # Reload that inventory, add another item, and save again.
    manager_reborn = ResourceManager(inventory)
    manager_reborn.inventory["added"] = {"name": "added",
                                         "type": "added-type",
                                         "id": "added-id"}
    manager_reborn._save()
    with open(inventory) as fh:
        content_reread = fh.read()
    for line in content:
        assert line in content_reread
    assert "added" in content_reread


def test_get_resources():
    manager = ResourceManager()
    manager.inventory = {
        "myshell": {"name": "myshell",
                    "type": "shell",
                    "id": "myshell-id"},
        "ambig-id0": {"name": "ambig-id0",
                      "type": "shell",
                      "id": "ambig-id"},
        "ambig-id1": {"name": "ambig-id1",
                      "type": "shell",
                      "id": "ambig-id"},
        "id-name-same": {"name": "id-name-same",
                         "type": "shell",
                         "id": "uniq-id"},
        "same-id": {"name": "same-id",
                    "type": "shell",
                    "id": "id-name-same"}
    }

    with pytest.raises(ResourceError):
        manager.get_resource("not there")

    resource_uniq = manager.get_resource("myshell")
    assert resource_uniq.name == "myshell"
    # We can get the same resource by ID.
    assert resource_uniq.id == "myshell-id"

    with pytest.raises(MultipleResourceMatches):
        manager.get_resource("ambig-id")

    # We get an exception if both if there is an name-ID collision...
    with pytest.raises(MultipleResourceMatches):
        manager.get_resource("id-name-same")
    # ... but we can disambiguate with resref_type.
    assert manager.get_resource("id-name-same", "name").id == "uniq-id"
    assert manager.get_resource("id-name-same", "id").id == "id-name-same"


def test_create_conflict():
    manager = ResourceManager()
    manager.inventory = {"already-exists": {"name": "already-exists",
                                            "type": "shell"}}
    with pytest.raises(ResourceAlreadyExistsError):
        manager.create("already-exists", "type-doesnt-matter")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
import logging
import pytest
import tempfile
import uuid

from ...cmd import Runner
from ...distributions.redhat import RPMTracer
from ...distributions.redhat import RPMPackage
from ...distributions.redhat import RedhatDistribution
from ...formats import Provenance
from ...resource.docker_container import DockerContainer
from ...tests.utils import skip_if_no_network, skip_if_no_docker_engine
from ...utils import swallow_logs


@skip_if_no_docker_engine
@pytest.fixture(scope='function')
def docker_container():
    skip_if_no_network()
    name = str(uuid.uuid4())  # Generate a random name for the container.
    Runner().run(['docker', 'run', '-t', '-d', '--rm', '--name',
        name, 'centos:7'])
    yield name
    Runner().run(['docker', 'stop', name])


@pytest.fixture(scope='module')
def centos_spec():
    tmp = tempfile.NamedTemporaryFile(delete=False)
    with open(tmp.name, 'w') as f:
        f.write("""# NICEMAN Environment Configuration File
# This file was created by NICEMAN 0.0.5 on 2018-05-23 22:03:22.820177
version: 0.0.1
distributions:
- name: redhat
  sources:
  - id: base/7/x86_64
    name: CentOS-7 - Base
    revision: '1525378614'
    updated: Thu May  3 20:17:37 2018
    mirrors: http://mirrorlist.centos.org/?release=7&arch=x86_64&repo=os&infra=container
    baseurl: http://mirror.wdc1.us.leaseweb.net/centos/7.5.1804/os/x86_64/ (9
    expire: '21600 second(s) (last: Wed May 23 22:02:41 2018)'
    filename: /etc/yum.repos.d/CentOS-Base.repo
  - id: epel/x86_64
    name: Extra Packages for Enterprise Linux 7 - x86_64
    revision: '1527088265'
    updated: Wed May 23 15:25:37 2018
    metalink: https://mirrors.fedoraproject.org/metalink?repo=epel-7&arch=x86_64
    baseurl: http://mirror.math.princeton.edu/pub/epel/7/x86_64/ (35 more)
    expire: '21600 second(s) (last: Wed May 23 22:02:43 2018)'
    filename: /etc/yum.repos.d/epel.repo
  - id: extras/7/x86_64
    name: CentOS-7 - Extras
    revision: '1526808850'
    updated: Sun May 20 09:36:03 2018
    mirrors: http://mirrorlist.centos.org/?release=7&arch=x86_64&repo=extras&infra=container
    baseurl: http://www.gtlib.gatech.edu/pub/centos/7.5.1804/extras/x86_64/ (9
    expire: '21600 second(s) (last: Wed May 23 22:02:43 2018)'
    filename: /etc/yum.repos.d/CentOS-Base.repo
  - id: updates/7/x86_64
    name: CentOS-7 - Updates
    revision: '1527083651'
    updated: Wed May 23 13:58:10 2018
    mirrors: http://mirrorlist.centos.org/?release=7&arch=x86_64&repo=updates&infra=container
    baseurl: http://mirror.math.princeton.edu/pub/centos/7.5.1804/updates/x86_64/
    expire: '21600 second(s) (last: Wed May 23 22:02:43 2018)'
    filename: /etc/yum.repos.d/CentOS-Base.repo
  packages:
  - name: fido
    pkgid: fido-1.1.2-1.el7.x86_64
    version: 1.1.2
    release: 1.el7
    architecture: x86_64
    install_date: Wed May 23 22:03:00 2018
    group: System Environment/Daemons
    size: '116736'
    license: GPLv2+ and LGPLv2+
    signature: RSA/SHA256, Wed Sep 24 16:10:18 2014, Key ID 6a2faea2352c64e5
    source_rpm: fido-1.1.2-1.el7.src.rpm
    build_date: Tue Sep 23 20:23:46 2014
    build_host: buildvm-17.phx2.fedoraproject.org
    packager: Fedora Project
    vendor: Fedora Project
    url: http://www.joedog.org/fido-home/
    files:
    - /usr/sbin/fido
  version: CentOS Linux release 7.4.1708 (Core)
""")
    return tmp.name


@pytest.fixture
def setup_packages():
    """set up the package comparison tests"""
    a = RPMPackage(name='p1')
    b = RPMPackage(name='p1', version='1.0')
    c = RPMPackage(name='p1', version='1.1')
    d = RPMPackage(name='p1', architecture='i386')
    e = RPMPackage(name='p1', architecture='alpha')
    f = RPMPackage(name='p1', version='1.1', architecture='i386')
    g = RPMPackage(name='p2')
    return (a, b, c, d, e, f, g)


@pytest.fixture
def setup_distributions(setup_packages):
    (p1, p1v10, p1v11, p1ai, p1aa, p1v11ai, p2) = setup_packages
    d1 = RedhatDistribution(name='debian 1')
    d1.packages = [p1]
    d2 = RedhatDistribution(name='debian 2')
    d2.packages = [p1v11]
    return (d1, d2)


def test_package_satisfies(setup_packages):
    (p1, p1v10, p1v11, p1ai, p1aa, p1v11ai, p2) = setup_packages
    assert p1.satisfies(p1)
    assert p1v10.satisfies(p1v10)
    assert not p1.satisfies(p1v10)
    assert p1v10.satisfies(p1)
    assert not p1v10.satisfies(p1v11)
    assert not p1.satisfies(p2)
    assert not p1v10.satisfies(p2)
    assert not p2.satisfies(p1v10)
    assert not p1v10.satisfies(p1aa)
    assert p1aa.satisfies(p1)
    assert not p1aa.satisfies(p1v10)
    assert not p1aa.satisfies(p1ai)
    assert not p1v11.satisfies(p1v11ai)
    assert p1v11ai.satisfies(p1v11)


def test_distribution_satisfies_package(setup_distributions, setup_packages):
    (d1, d2) = setup_distributions
    (p1, p1v10, p1v11, p1ai, p1aa, p1v11ai, p2) = setup_packages
    assert d1.satisfies_package(p1)
    assert not d1.satisfies_package(p1v10)
    assert d2.satisfies_package(p1)
    assert not d2.satisfies_package(p1v10)
    assert d2.satisfies_package(p1v11)


def test_distribution_statisfies(setup_distributions):
    (d1, d2) = setup_distributions
    assert not d1.satisfies(d2)
    assert d2.satisfies(d1)


def test_distribution_sub(setup_packages):
    (p1, p1v10, p1v11, p1ai, p1aa, p1v11ai, p2) = setup_packages
    d1 = RedhatDistribution(name='debian 1')
    d1.packages = [p1, p2]
    d2 = RedhatDistribution(name='debian 2')
    d2.packages = [p1v11, p2]
    assert d1-d2 == []
    result = d2-d1
    assert len(result) == 1
    assert result[0] == p1v11


@skip_if_no_docker_engine
def test_tracer(docker_container):
    skip_if_no_network()

    # Test setup
    resource = DockerContainer(docker_container)
    resource.connect()
    session = resource.get_session()
    tracer = RPMTracer(session)
    tracer._init()

    # Test sorting through a given set of files and identifying the ones
    # that belong to CentOS.
    files = ['/usr/bin/ls', 'foo']
    dist, remaining_files = next(tracer.identify_distributions(files))
    assert 'foo' in remaining_files
    assert dist.name == 'redhat'
    assert dist.version.startswith('CentOS Linux release 7')
    source_ids = [s.id for s in dist.sources]
    assert 'base/7/x86_64' in source_ids
    assert 'extras/7/x86_64' in source_ids
    assert 'updates/7/x86_64' in source_ids
    assert dist.packages[0].name == 'coreutils'
    assert dist.packages[0].group == 'System Environment/Base'
    assert dist.packages[0].files[0] == '/usr/bin/ls'

    # Test creating RPMPackage objects with files found to be in CentOS.
    packages = tracer._get_packagefields_for_files(['/usr/bin/ls'])
    assert packages['/usr/bin/ls']['name'] == 'coreutils'
    assert packages['/usr/bin/ls']['packager'].startswith('CentOS BuildSystem')


@skip_if_no_docker_engine
def test_distribution(docker_container, centos_spec):
    skip_if_no_network()

    # Test setup
    resource = DockerContainer(docker_container)
    resource.connect()
    session = resource.get_session()
    provenance = Provenance.factory(centos_spec, 'niceman')
    dist = provenance.get_distributions()[0]

    with swallow_logs(new_level=logging.DEBUG) as log:
        dist.initiate(session)
        assert "Running command ['yum', 'repolist']" in log.lines
        assert "Running command ['yum', 'install', '-y', 'epel-release']" in \
            log.lines

        dist.install_packages(session)
        assert "Installing fido-1.1.2-1.el7.x86_64" in log.lines
        assert "Running command ['yum', 'install', '-y', 'fido-1.1.2-1.el7.x86_64']" \
            in log.lines



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Common tests configuration for py.test
"""

# Some commonly used fixtures

from niceman.formats.tests.fixtures import demo1_spec, reprozip_spec2

import pytest


def pytest_addoption(parser):
    parser.addoption("--integration", action="store_true",
                     default=False, help="run integration tests")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--integration"):
        # --integration given in cli: do not skip integration tests
        return
    skip_integration = pytest.mark.skip(
        reason="need --integration option to run")
    for item in items:
        if "integration" in item.keywords:
            item.add_marker(skip_integration)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Analyze existing spec or session file system to gather more detailed information
"""

from __future__ import unicode_literals

from os.path import normpath
import sys
import time

from niceman.resource.session import get_local_session
from niceman.resource.session import Session
from .common_opts import resref_opt
from .common_opts import resref_type_opt
from .base import Interface
from ..support.constraints import EnsureNone
from ..support.constraints import EnsureStr
from ..support.exceptions import InsufficientArgumentsError
from ..support.param import Parameter
from ..utils import assure_list
from ..utils import to_unicode
from ..resource import get_manager

__docformat__ = 'restructuredtext'

from logging import getLogger
lgr = getLogger('niceman.api.retrace')


class Retrace(Interface):
    """Gather detailed package information from paths or a ReproZip trace file.

    Examples
    --------

      $ niceman retrace --spec reprozip_run.yml > niceman_config.yml

    """

    _params_ = dict(
        spec=Parameter(
            args=("--spec",),
            doc="ReproZip YML file to be analyzed",
            metavar='SPEC',
            # nargs="+",
            constraints=EnsureStr() | EnsureNone(),
        ),
        path=Parameter(
            args=("path",),
            metavar="PATH",
            doc="""path(s) to be traced.  If spec is provided, would trace them
            after tracing the spec""",
            nargs="*",
            constraints=EnsureStr() | EnsureNone()),
        output_file=Parameter(
            args=("-o", "--output-file",),
            doc="Output file.  If not specified - printed to stdout",
            metavar='output_file',
            constraints=EnsureStr() | EnsureNone(),
        ),
        resref=Parameter(
            args=("-r", "--resource",),
            dest="resref",
            metavar="RESOURCE",
            doc="""Name or ID of the resource to operate on. To see available
            resources, run 'niceman ls'.[PY: Note: As a special case, a session
            instance can be passed as the value for `resref`.  PY]""",
            constraints=EnsureStr() | EnsureNone()),
        resref_type=resref_type_opt,
    )

    # TODO: add a session/resource so we could trace within
    # arbitrary sessions
    @staticmethod
    def __call__(path=None, spec=None, output_file=None,
                 resref=None, resref_type="auto"):
        # heavy import -- should be delayed until actually used

        if not (spec or path):
            raise InsufficientArgumentsError(
                "Need at least a single --spec or a file"
            )

        paths = assure_list(path)
        if spec:
            lgr.info("reading spec file %s", spec)
            # TODO: generic loader to auto-detect formats etc
            from niceman.formats.reprozip import ReprozipProvenance
            spec = ReprozipProvenance(spec)
            paths += spec.get_files() or []

        # Convert paths to unicode
        paths = map(to_unicode, paths)
        # The tracers assume normalized paths.
        paths = list(map(normpath, paths))

        if isinstance(resref, Session):
            # TODO: Special case for Python callers.  Is this something we want
            # to handle more generally at the interface level?
            session = resref
        elif resref:
            resource = get_manager().get_resource(resref, resref_type)
            session = resource.get_session()
        else:
            session = get_local_session()

        # TODO: at the moment assumes just a single distribution etc.
        #       Generalize
        # TODO: RF so that only the above portion is reprozip specific.
        # If we are to reuse their layout largely -- the rest should stay as is
        (distributions, files) = identify_distributions(
            paths,
            session=session
        )
        from niceman.distributions.base import EnvironmentSpec
        spec = EnvironmentSpec(
            distributions=distributions,
        )
        if files:
            spec.files = sorted(files)

        # TODO: generic writer!
        from niceman.formats.niceman import NicemanProvenance
        stream = open(output_file, "w") if output_file else sys.stdout
        NicemanProvenance.write(stream, spec)
        if stream is not sys.stdout:
            stream.close()


# TODO: session should be with a state.  Idea is that if we want
#  to trace while inheriting all custom PATHs which that run might have
#  had
def identify_distributions(files, session=None, tracer_classes=None):
    """Identify packages files belong to

    Parameters
    ----------
    files : iterable
      Files to consider

    Returns
    -------
    distributions : list of Distribution
    unknown_files : list of str
      Files which were not determined to belong to any specific distribution
    """
    if tracer_classes is None:
        tracer_classes = get_tracer_classes()

    session = session or get_local_session()
    # TODO create list of appropriate for the `environment` OS tracers
    #      in case of no environment -- get current one
    # TODO: should operate in the session, might be given additional information
    #       not just files


    # .identify_ functions will have a side-effect of shrinking this list in-place
    # as they identify files beloning to them
    files_to_consider = set(files)

    distibutions = []
    files_processed = set()
    files_to_trace = files_to_consider

    niter = 0
    max_niter = 10
    while True:
        niter += 1
        nfiles_processed = len(files_processed)
        nfiles_to_trace = len(files_to_trace)
        lgr.info("Entering iteration #%d over Tracers", niter)
        if niter > max_niter:
            lgr.error(
                "We did %s iterations already, something is not right"
                % max_niter)
            break

        for Tracer in tracer_classes:
            lgr.debug("Tracing using %s", Tracer.__name__)
            # TODO: memoize across all loops
            # Identify directories from the files_to_consider
            dirs = set(filter(session.isdir, files_to_trace))

            # Pull out directories if the tracer can't handle them
            if Tracer.HANDLES_DIRS:
                files_to_trace = files_to_consider
                files_skipped = set()
            else:
                files_to_trace = files_to_consider - dirs
                files_skipped = files_to_consider - files_to_trace

            tracer = Tracer(session=session)
            begin = time.time()
            # yoh things the idea was that tracer might trace even without
            #     files, so we should not just 'continue' the loop if there is no
            #     files_to_trace
            if files_to_trace:
                remaining_files_to_trace = files_to_trace
                nenvs = 0
                for env, remaining_files_to_trace in tracer.identify_distributions(
                        files_to_trace):
                    distibutions.append(env)
                    nenvs += 1
                files_processed |= files_to_trace - remaining_files_to_trace
                files_to_trace = remaining_files_to_trace
                lgr.info("%s: %d envs with %d other files remaining",
                         Tracer.__name__,
                         nenvs,
                         len(files_to_trace))

            # Re-combine any files that were skipped
            files_to_consider = files_to_trace | files_skipped

            lgr.debug("Assigning files to packages by %s took %f seconds",
                      tracer, time.time() - begin)
        if len(files_to_trace) == 0 or (
            nfiles_processed == len(files_processed) and
            nfiles_to_trace == len(files_to_trace)):
            lgr.info("No more changes or files to track.  Exiting the loop")
            break

    return distibutions, files_to_consider


def get_tracer_classes():
    """A helper which returns a list of all available Tracers

    The order should not but does matter and ATM is magically provided
    """
    # TODO: automate discovery of available tracers
    from niceman.distributions.debian import DebTracer
    from niceman.distributions.redhat import RPMTracer
    from niceman.distributions.conda import CondaTracer
    from niceman.distributions.venv import VenvTracer
    from niceman.distributions.vcs import VCSTracer
    from niceman.distributions.docker import DockerTracer
    from niceman.distributions.singularity import SingularityTracer
    Tracers = [DebTracer, RPMTracer, CondaTracer, VenvTracer, VCSTracer,
        DockerTracer, SingularityTracer]
    return Tracers



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##


# NOTE: The singularity classes SingularitySession and PTYSingularitySession
# are tested in test_session.test_session_abstract_methods()

import os
import pytest
import tempfile

import logging
import re
from ..singularity import Singularity, SingularitySession
from ...tests.utils import skip_if_no_singularity, skip_if_no_network, \
    swallow_logs, assert_in


@skip_if_no_network
@skip_if_no_singularity
def test_singularity_resource_class():

    # Set working directory to a scratch directory since we will be creating
    # Singularity image files during testing.
    orig_cwd = os.getcwd()
    tempdir = tempfile.mkdtemp()
    os.chdir(tempdir)

    with swallow_logs(new_level=logging.DEBUG) as log:

        # Test creating a new singularity container instance.
        resource = Singularity(name='foo',
            image='shub://truatpasteurdotfr/singularity-alpine')
        assert resource.name == 'foo'
        assert resource.image == 'shub://truatpasteurdotfr/singularity-alpine'
        resource.connect()
        assert resource.id is None
        assert resource.status is None
        resource.create()
        assert resource.id.startswith('foo-')
        assert resource.status == 'running'

        # Test trying to create an already running instance.
        resource_duplicate = Singularity(name='foo',
            image='shub://truatpasteurdotfr/singularity-alpine')
        resource_duplicate.connect()
        assert resource_duplicate.id.startswith('foo-')
        assert resource_duplicate.status == 'running'
        resource_duplicate.create()
        assert_in('Resource foo already exists.', log.lines)

        # Test retrieving instance info.
        info = resource.get_instance_info()
        assert info['name'] == 'foo'
        assert re.match(r'^\d+$', info['pid'])
        assert info['image'].endswith('.simg')

        # Test starting an instance.
        with pytest.raises(NotImplementedError):
            resource.start()

        # Test stopping an instance.
        with pytest.raises(NotImplementedError):
            resource.stop()

        # Test getting a resource session.
        session = resource.get_session()
        assert isinstance(session, SingularitySession)

        # Test deleting an instance.
        resource.delete()
        assert resource.id is None
        assert resource.status is None

        # Test retrieving info from a non-existent instance.
        info = resource.get_instance_info()
        assert info is None

    # Return to original working directory
    os.chdir(orig_cwd)


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Plugin support for provenance YAML files produced by ReproZip utility.

See: https://vida-nyu.github.io/reprozip/
"""

import io
import yaml

from .base import Provenance

import logging
lgr = logging.getLogger('niceman.formats.reprozip')


class ReprozipProvenance(Provenance):
    """Parser for ReproZip file format (YAML specification) """

    @classmethod
    def _load(cls, source):
        with io.open(source, encoding='utf-8') as stream:
            config = yaml.safe_load(stream)
            # TODO: Check version of ReproZip file and warn if unknown
            return config

    # Might come handy to define 'base' whenever we get there
    # def get_os(self):
    #     return self.yaml['runs'][0]['distribution'][0]
    #
    # def get_os_version(self):
    #     return self.yaml['runs'][0]['distribution'][1]
    #
    # def get_create_date(self):
    #     format = '%Y%m%dT%H%M%SZ'
    #     return self.yaml['runs'][0]['date'].strftime(format)
    #
    # def get_environment_vars(self):
    #     return self.yaml['runs'][0]['environ']
    #
    # def get_packages(self):
    #     return [{'name': p['name'], 'version': p['version']} for p in self.yaml['packages']]
    #
    # def get_commandline(self):
    #     return self.yaml['runs'][0]['argv']

    def get_files(self, limit='all'):
        """Pulls the system files from a ReproZip configuration into a set
    
        Given a ReproZip configuration (read into a dictionary) it pulls
        the list of files from "packages" and "other files" sections into a
        set. It excludes files from "input_output".
    
        Parameters
        ----------
        config : dict
            ReproZip configuration
        other_files : bool, optional
            Either to return also other_files
    
        Return
        ------
        set
            Files listed in the configuration
        """

        files = set()

        src_yaml = self._src
        if limit in {'all', 'packaged'}:
            for package in src_yaml.get('packages', []) or []:
                if 'files' in package:
                    files.update(package.get('files', []))

        if limit in {'all', 'loose'} and 'other_files' in src_yaml:
            files.update(src_yaml.get("other_files", []))

        return files



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from niceman.cmdline.main import main

import logging
from mock import patch, call, MagicMock

from niceman.utils import swallow_logs
from niceman.resource.base import ResourceManager
from niceman.tests.utils import assert_in


def test_login_interface():
    """
    Test logging into an environment
    """

    with patch('docker.Client') as client, \
        patch('niceman.resource.ResourceManager._get_inventory') as get_inventory, \
        patch('dockerpty.start'), \
        swallow_logs(new_level=logging.DEBUG) as log:

        client.return_value = MagicMock(
            containers=lambda all: [
                {
                    'Id': '18b31b30e3a5',
                    'Names': ['/my-test-resource'],
                    'State': 'running'
                }
            ],
        )

        get_inventory.return_value = {
            "my-test-resource": {
                "status": "running",
                "engine_url": "tcp://127.0.0.1:2375",
                "type": "docker-container",
                "name": "my-test-resource",
                "id": "18b31b30e3a5"
            }
        }

        args = ['login',
                'my-test-resource'
        ]

        with patch("niceman.interface.login.get_manager",
                   return_value=ResourceManager()):
            main(args)

        assert client.call_count == 1

        calls = [
            call(base_url='tcp://127.0.0.1:2375')
        ]
        client.assert_has_calls(calls, any_order=True)

        assert_in("Opening TTY connection to docker container.", log.lines)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from ..retrace import identify_distributions


# TODO: mock it up or run only when under Debian
def test_identify_packages():
    files = ["/usr/share/doc/xterm/copyright",
             "/usr/games/alienblaster",
             "/usr/share/icons/hicolor/48x48/apps/xterm-color.png",
             "/usr/share/doc/zlib1g/copyright",
             "/usr/bin/vim.basic",
             "/usr/share/bug/vim/script",
             "/home/butch"]
    # Simple sanity check that the pipeline works
    distributions, files = identify_distributions(files)
    assert True



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
import os
import os.path as op
import sys

from appdirs import AppDirs
import attr
import pytest
import logging

from niceman.cmd import Runner
from niceman.utils import chpwd
from niceman.tests.utils import create_pymodule
from niceman.tests.utils import skip_if_no_network, assert_is_subset_recur
from niceman.tests.utils import swallow_logs
from niceman.distributions.venv import VenvDistribution
from niceman.distributions.venv import VenvEnvironment
from niceman.distributions.venv import VenvPackage
from niceman.distributions.venv import VenvTracer

PY_VERSION = "python{v.major}.{v.minor}".format(v=sys.version_info)


@pytest.fixture(scope="session")
@skip_if_no_network
def venv_test_dir():
    dirs = AppDirs('niceman')
    test_dir = os.path.join(dirs.user_cache_dir, 'venv_test')
    if os.path.exists(test_dir):
        return test_dir

    runner = Runner()
    runner.run(["mkdir", "-p", test_dir])

    pymod_dir = os.path.join(test_dir, "minimal_pymodule")
    create_pymodule(pymod_dir)

    with chpwd(test_dir):
        runner.run(["virtualenv", "--python", PY_VERSION, "venv0"])
        runner.run(["virtualenv", "--python", PY_VERSION, "venv1"])
        runner.run(["./venv0/bin/pip", "install", "pyyaml"])
        runner.run(["./venv0/bin/pip", "install", "-e", pymod_dir])
        runner.run(["./venv1/bin/pip", "install", "attrs"])
        # Make sure we're compatible with older pips.
        runner.run(["./venv1/bin/pip", "install", "pip==9.0.3"])
    return test_dir


@pytest.mark.integration
def test_venv_identify_distributions(venv_test_dir):
    paths = ["lib/" + PY_VERSION + "/site-packages/yaml/parser.py",
             "lib/" + PY_VERSION + "/site-packages/attr/filters.py"]

    with chpwd(venv_test_dir):
        path_args = [
            # Both full ...
            os.path.join(venv_test_dir, "venv0", paths[0]),
            # ... and relative paths work.
            os.path.join("venv1", paths[1]),
        ]
        path_args.append("/sbin/iptables")

        tracer = VenvTracer()

        dists = list(tracer.identify_distributions(path_args))
        assert len(dists) == 1

        distributions, unknown_files = dists[0]
        assert unknown_files == {
            "/sbin/iptables",
            # The editable package was added by VenvTracer as an unknown file.
            os.path.join(venv_test_dir, "minimal_pymodule")}

        assert len(distributions.environments) == 2

        expect = {"environments":
                  [{"packages": [{"files": [paths[0]], "name": "PyYAML",
                                  "editable": False},
                                 {"files": [], "name": "nmtest",
                                  "editable": True}]},
                   {"packages": [{"files": [paths[1]], "name": "attrs",
                                  "editable": False}]}]}
        assert_is_subset_recur(expect, attr.asdict(distributions), [dict, list])


@pytest.mark.integration
def test_venv_install(venv_test_dir, tmpdir):
    tmpdir = str(tmpdir)
    paths = ["lib/" + PY_VERSION + "/site-packages/yaml/parser.py",
             "lib/" + PY_VERSION + "/site-packages/attr/filters.py"]

    tracer = VenvTracer()
    dists = list(
        tracer.identify_distributions(
            [op.join(venv_test_dir, "venv0", paths[0]),
             op.join(venv_test_dir, "venv1", paths[1])]))
    assert len(dists) == 1
    dist = dists[0][0]

    assert len(dist.environments) == 2
    for idx in range(2):
        dist.environments[idx].path = op.join(tmpdir, "venv{}".format(idx))

    dist.initiate(None)
    dist.install_packages()

    dists_installed = list(
        tracer.identify_distributions(
            [op.join(tmpdir, "venv0", paths[0]),
             op.join(tmpdir, "venv1", paths[1])]))
    assert len(dists_installed) == 1
    dist_installed = dists_installed[0][0]

    expect = {"environments":
              [{"packages": [{"name": "PyYAML",
                              "editable": False}]},
               {"packages": [{"name": "attrs",
                              "editable": False}]}]}
    assert_is_subset_recur(expect, attr.asdict(dist_installed), [dict, list])

    # We don't yet handle editable packages.
    assert any([p.name == "nmtest"
                for e in dist.environments
                for p in e.packages])
    assert not any([p.name == "nmtest"
                    for e in dist_installed.environments
                    for p in e.packages])


def test_venv_install_noop():
    dist = VenvDistribution(
        name="venv",
        path="/tmp/doesnt/matter/",
        venv_version="15.1.0",
        environments=[
            VenvEnvironment(
                path="/tmp/doesnt/matter/venv",
                python_version="3.7",
                packages=[
                    VenvPackage(
                        name="imeditable",
                        version="0.1.0",
                        editable=True,
                        local=True)])])
    with swallow_logs(new_level=logging.INFO) as log:
        dist.install_packages()
        assert "No local, non-editable packages found" in log.out



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Classes to identify VCS repos for files"""

from __future__ import unicode_literals

import abc
import attr
import os

from collections import defaultdict
from os.path import dirname, isdir, isabs, abspath
from os.path import exists, lexists
from os.path import join as opj

from logging import getLogger

from niceman.dochelpers import exc_str
from niceman.utils import attrib
from niceman.utils import only_with_values
from niceman.utils import instantiate_attr_object
from niceman.resource.session import get_local_session

from niceman.cmd import CommandError, Runner

lgr = getLogger('niceman.distributions.vcs')

from niceman.distributions.base import DistributionTracer
from niceman.distributions.base import SpecObject
from niceman.distributions.base import Distribution
from niceman.distributions.base import TypedList


# # TODO: use metaclass I guess... ?
# def get_vcs_distribution(RepoClass, name, Name):
#     """A helper to generate VCS distribution classes"""
#     class VCSDistribution(Distribution):
#
#         repositories = TypedList(RepoClass)
#
#         def __init__(self, *args, **kwargs):
#             self.name = name
#             super(VCSDistribution, self).__init__(*args, **kwargs)
#     VCSDistribution.__name__ = str('%sDistribution' % Name)
#     return VCSDistribution


# We need a concept of Distribution
@attr.s
class VCSDistribution(Distribution):
    """
    Base class for VCS "distributions"
    """

    def initiate(self, session):
        # This is VCS specific, but we could may be make it
        # to verify that some executable is available
        session.execute_command(self._cmd)

    @abc.abstractmethod
    def install_packages(self, session, use_version=True):
        # This is VCS specific
        return

    def normalize(self):
        pass


@attr.s
class VCSRepo(SpecObject):
    """Base VCS repo class"""

    path = attrib(default=attr.NOTHING)
    files = attrib(default=attr.Factory(list))

    @property
    def identifier(self):
        try:
            return getattr(self, self._identifier_attribute)
        except AttributeError:
            # raised if _identifier_attribute is not defined, but this means
            # (to the caller) that identifier is not defined
            msg = "%s instance has no attribute 'identifier'" % self.__class__
            raise AttributeError(msg)

    @property
    def commit(self):
        try:
            return getattr(self, self._commit_attribute)
        except AttributeError:
            # as for .identifer above, reraise with a more appropriate message
            msg = "%s instance has no attribute 'commit'" % self.__class__
            raise AttributeError(msg)

@attr.s
class GitRepo(VCSRepo):

    root_hexsha = attrib()
    branch = attrib()
    hexsha = attrib()
    describe = attrib()
    tracked_remote = attrib()
    remotes = attrib(default=attr.Factory(dict))

    _identifier_attribute = 'root_hexsha'
    _commit_attribute = 'hexsha'

# Probably generation wouldn't be flexible enough
#GitDistribution = get_vcs_distribution(GitRepo, 'git', 'Git')
@attr.s
class GitDistribution(VCSDistribution):
    _cmd = "git"
    packages = TypedList(GitRepo)

    def initiate(self, session=None):
        pass

    def install_packages(self, session=None, use_version=True):
        session = session or get_local_session()
        for repo in self.packages:
            self._install_repo(session, repo)

    def _install_repo(self, session, repo):
        sources = {k: v for k, v in repo.remotes.items() if v.get("contains")}
        if not sources:
            lgr.warning("No remote known for '%s'; skipping", repo.path)
            return

        if repo.tracked_remote in sources:
            remote = repo.tracked_remote
        elif "origin" in sources:
            remote = "origin"
        else:
            # Just grab any remote at this point.  It doesn't really matter as
            # long as we get the target commit.
            remote = next(iter(sources.keys()))

        if session.exists(repo.path):
            cloned = False
            shim = self._get_matching_shim(session, repo)
            if shim is None:
                return
            if not shim.has_revision(repo.hexsha):
                shim._run_git(["fetch", remote])
        else:
            cloned = True
            clone_url = sources[remote]["url"]

            lgr.info("Cloning %s from %s (%s)", repo.path, clone_url, remote)
            session.execute_command(
                ["git", "clone", "-o", remote, clone_url, repo.path])
            shim = GitRepoShim.get_at_dirpath(session, repo.path)

        if repo.remotes:
            lgr.info("Adding remotes to %s", repo.path)
            current_remotes = set(shim._run_git(["remote"]).splitlines())
            for remote, remote_info in repo.remotes.items():
                if remote not in current_remotes:
                    try:
                        shim._run_git(["remote", "add", "-f",
                                       remote, remote_info["url"]])
                    except CommandError:
                        lgr.warning("Failed to fetch remote %s at %s",
                                    remote, remote_info["url"])

        if not shim.has_revision(repo.hexsha):
            lgr.warning("Set up '%s', but the expected hexsha wasn't found",
                        repo.path)
            return

        # Be less aggressive about restoring the revision/branch state if we
        # didn't clone the repo.
        self._checkout(shim, repo, force=cloned)

    @staticmethod
    def _get_matching_shim(session, repo):
        """Return a shim for the repository at `repo.path`.

        Parameters
        ----------
        session : Session object
        repo : GitRepo object

        Returns
        -------
        A GitRepoShim object for the repository at `repo.path`.  If `repo.path`
        is incompatible with some detail of `repo` (e.g., it is not a Git
        repository), a warning is issued and no value is returned.
        """
        if not session.isdir(repo.path):
            lgr.warning("'%s' is not a directory; skipping",
                        repo.path)
            return

        shim = GitRepoShim.get_at_dirpath(session, repo.path)
        if shim is None:
            # TODO: We could proceed if the directory is empty.
            lgr.warning("Directory '%s' exists, "
                        "but is not a Git repository; skipping",
                        repo.path)
            return
        if shim.root_hexsha != repo.root_hexsha:
            lgr.warning("Root hexsha in '%s' doesn't match "
                        "expected hexsha; skipping",
                        repo.path)
            return
        return shim

    @staticmethod
    def _checkout(shim, repo, force=False):
        """Try to checkout the recorded revision and branch.

        Parameters
        ----------
        shim : GitRepoShim object
            The new repository being set up.
        repo : GitRepo
            The recorded information being used to set up the new repository.
        force : bool
            Reset the branch in `shim` to `repo.hexsha` even if this branch
            already exists in the new repository.
        """

        if shim._run_git(["status", "--porcelain"]).strip():
            lgr.warning("Not setting HEAD to %s because repository is dirty",
                        repo.hexsha)
            return

        set_tracking = False
        if not repo.branch:
            checkout_args = [repo.hexsha]
        elif repo.branch == shim.branch and shim.hexsha == repo.hexsha:
            checkout_args = None
        elif force:
            checkout_args = ["-B", repo.branch, repo.hexsha]
            set_tracking = True
        elif repo.branch == shim.branch:
            checkout_args = [repo.hexsha]
        else:
            heads = dict(
                name_hexsha.split("\0") for name_hexsha in
                shim._run_git(["for-each-ref", "--format",
                               "%(refname:short)%00%(objectname)",
                               "refs/heads"]).splitlines())
            if repo.branch in heads and heads[repo.branch] == repo.hexsha:
                checkout_args = [repo.branch]
                set_tracking = True
            else:
                checkout_args = [repo.hexsha]

        if checkout_args is not None:
            shim._run_git(["checkout"] + checkout_args)
            if set_tracking and repo.tracked_remote:
                for var, value in [("remote", repo.tracked_remote),
                                   # Note: We're assuming that the traced
                                   # branch has the same name.  GitRepo doesn't
                                   # actually store this information.
                                   ("merge", "refs/heads/" + repo.branch)]:
                    shim._run_git(["config",
                                   ".".join(["branch", repo.branch, var]),
                                   value])


GitRepo._distribution = GitDistribution


@attr.s
class SVNRepo(VCSRepo):

    revision = attrib()
    url = attrib()
    root_url = attrib()
    relative_url = attrib()
    uuid = attrib()

    _identifier_attribute = 'uuid'
    _commit_attribute = 'revision'

#SVNDistribution = get_vcs_distribution(SVNRepo, 'svn', 'SVN')
@attr.s
class SVNDistribution(VCSDistribution):
    _cmd = "svn"
    packages = TypedList(SVNRepo)

    def install_packages(self, session, use_version=True):
        raise NotImplementedError
SVNRepo._distribution = SVNDistribution


#
# Tracer Shims
# We use unified VCSTracer but it needs per-VCS specific handling/
# VCS objects are not created to worry/carry the session information
# so we have per-VCS shim objects which would be tracer helpers
#
class GitSVNRepoShim(object):
    _ls_files_command = None  # just need to define in subclass
    _ls_files_filter = None

    _vcs_class = None  # associated VCS class

    def __init__(self, path, session):
        """Representation for a repository

        Parameters
        ----------
        path: str
           Path to the top of the repository

        """
        self.path = path.rstrip(os.sep)  # TODO: might be done as some rg to attr.ib
        self._session = session
        self._all_files = None
        self._branch = None

    def _session_execute_command(self, cmd, **kwargs):
        """Run in the session but providing our self.path as the cwd"""
        if 'cwd' not in kwargs:
            kwargs = dict(cwd=self.path, **kwargs)
        return self._session.execute_command(cmd, **kwargs)

    @property
    def all_files(self):
        """Lazy evaluation for _all_files. If session changes, result would be old"""
        if self._all_files is None:
            out, err = self._session_execute_command(self._ls_files_command)
            assert not err
            self._all_files = set(filter(None, out.split('\n')))
            if self._ls_files_filter:
                self._all_files = self._ls_files_filter(self._all_files)
            self._all_files = set(self._all_files)  # for efficient lookups
        return self._all_files

    def owns_path(self, path):
        # does this repository have this directory under its control?
        if path.rstrip(os.sep) == self.path:
            return True
        # it might be that the path points to itself
        # NOTE:
        #  this is an interesting case -- don't we want to still track "possible"
        #  artifacts produced from the repos?  e.g. ./build subdirectory
        #  But for that we would need to ask twice -- first a full sweep across
        #  all Resolvers making it all strict checks, and then again through
        #  those which could potentially claim it to belong to the repo?
        #  For now just a strict check, and we would want to request all files
        #  which repo knows about
        rpath = path[len(self.path)+1:]
        return rpath in self.all_files

    @classmethod
    def get_at_dirpath(cls, session, dirpath):
        """Return VCS instance at the given path (if under that VCS control)"""
        raise NotImplementedError



# Name must be   TYPERepo since used later in the code
# As an overkill might want some metaclass to look after that ;-)
class SVNRepoShim(GitSVNRepoShim):

    _vcs_class = SVNRepo
    _vcs_distribution_class = SVNDistribution

    @property
    def _ls_files_command(self):
        # tricky -- we need to locate wc.db somewhere upstairs, and filter out paths
        root_path = self._info['Working Copy Root Path']
        return 'sqlite3 -noheader "%s/.svn/wc.db" ' \
            '"select local_relpath from nodes_base"' \
                    % root_path

    def _ls_files_filter(self, all_files):
        root_path = self._info['Working Copy Root Path']
        subdir = os.path.relpath(self.path, root_path)
        if subdir == os.curdir:
            return all_files
        else:
            return [f[len(subdir)+1:] for f in all_files
                    if os.path.commonprefix((subdir, f)) == subdir]

    def __init__(self, *args, **kwargs):
        super(SVNRepoShim, self).__init__(*args, **kwargs)
        self.__info = None

    @classmethod
    def get_at_dirpath(cls, session, dirpath):
        # ho ho -- no longer the case that there is .svn in each subfolder:
        # http://stackoverflow.com/a/9070242
        found = False
        if exists(opj(dirpath, '.svn')):  # early detection
            found = True
        # but still might be under SVN
        if not found:
            try:
                out, err = session.execute_command(
                    'svn info',
                    # expect_fail=True,
                    cwd=dirpath
                )
            except CommandError as exc:
                if "Please see the 'svn upgrade' command" in str(exc):
                    lgr.warning(
                        "SVN at %s is outdated, needs 'svn upgrade'",
                        dirpath)
                else:
                    # we are in SVN but it is outdated and needs an upgrade
                    lgr.debug(
                        "Probably %s is not under SVN repo path: %s",
                        dirpath, exc_str(exc)
                    )
                    return None
        # for now we treat each directory under SVN independently
        # pros:
        #   - could provide us 'minimal' set of checkouts to do, since it might be
        #      quite expensive to checkout the entire tree if it was not used
        #      besides few leaves
        lgr.debug("Detected SVN repository at %s", dirpath)
        return cls(dirpath, session=session)

    @property
    def _info(self):
        if self.__info is None:
            # TODO -- outdated repos might need 'svn upgrade' first
            # so not sure -- if we should copy them somewhere first and run
            # update there or ask user to update them on his behalf?!
            out, err = self._session.execute_command('svn info', cwd=self.path)
            self.__info = dict(
                [x.lstrip() for x in l.split(':', 1)]
                for l in out.splitlines() if l.strip()
            )
        return self.__info

    @property
    def revision(self):
        # svn info doesn't give the current revision
        # (see http://svnbook.red-bean.com/en/1.7/svn.tour.history.html)
        # so we need to run svn log
        if not hasattr(self, '_revision'):
            runner = Runner()
            log = runner(['svn', 'log', '^/', '-l', '1'], cwd=self.path)[0]
            lines = log.strip().split('\n')
            if len(lines) == 1:
                self._revision = None
            else:
                revision = lines[1].split()[0]
                if revision.startswith('r') and revision[1:].isdigit():
                    self._revision = int(revision[1:])
                else:
                    self._revision = revision[1:]
        return self._revision

    @property
    def url(self):
        # also has similarity to APT in that we could have the top of SVN repo
        # as an 'origin' which might be reused by multiple 'sub-repos'/directories
        # "Repository Root" and "Relative URL"
        return self._info['URL']

    @property
    def root_url(self):
        return self._info['Repository Root']

    @property
    def relative_url(self):
        return self._info['Relative URL']

    @property
    def uuid(self):
        return self._info['Repository UUID']


class GitRepoShim(GitSVNRepoShim):

    _ls_files_command = 'git ls-files'

    _vcs_class = GitRepo
    _vcs_distribution_class = GitDistribution

    @classmethod
    def get_at_dirpath(cls, session, dirpath):
        try:
            out, err = session.execute_command(
                'git rev-parse --show-toplevel',
                # expect_fail=True,
                cwd=dirpath
            )
        except CommandError as exc:
            lgr.debug(
                "Probably %s is not under git repo path: %s",
                dirpath, exc_str(exc)
            )
            return None
        topdir = out.rstrip('\n')
        lgr.debug("Detected Git repository at %s for %s. Creating a session shim", topdir, dirpath)
        return cls(topdir, session=session)

    def _run_git(self, cmd, expect_fail=False, **kwargs):
        """Helper to run git command, and ignore stderr"""
        cmd = ['git'] + cmd if isinstance(cmd, list) else 'git ' + cmd
        try:
            out, err = self._session.execute_command(
                cmd,
                #expect_fail=expect_fail,
                cwd=self.path,
                **kwargs)
        except CommandError:
            if not expect_fail:
                raise
            else:
                return None
        return out.strip()

    @property
    def hexsha(self):
        try:
            return self._run_git('rev-parse HEAD')
        except CommandError:
            # might still be the first yet to be committed state in the branch
            return None

    @property
    def root_hexsha(self):
        try:
            rev_list = self._run_git('rev-list --max-parents=0 HEAD').split('\n')
            return rev_list[-1]
        except CommandError:
            # might still be the first yet to be committed state in the branch
            return None

    @property
    def describe(self):
        """Let's use git describe"""
        try:
            return self._run_git('describe --tags', expect_fail=True)
        except CommandError:
            return None

    @property
    def remotes(self):
        # ideally needs to figure out the remote(s) which already have
        # this commit.  So kinda similar to APT where we list all origins
        # where some might not even contain current version. TODO: implement
        # that way
        #
        # On the other hand, I think it would not be uncommon to use some
        # version which is not yet pushed... so what additional information
        # would this check provide us?  We better record current branch,
        # and mark remote which is tracked for it
        hexsha = self.hexsha
        # which remotes contain this commit, so we could provide this
        # possibly valuable information
        if not hexsha:  # just initialized
            return {}

        remote_branches = self._run_git(
            ["branch", "-r", "--contains", hexsha]).splitlines()
                                               # e.g. "origin/HEAD -> origin/master"
        remote_branches = [b.strip() for b in remote_branches if " -> " not in b]

        if not remote_branches:
            return {}
        containing_remotes = set(x.split('/', 1)[0] for x in remote_branches)
        remotes = {}
        for remote in self._run_git('remote').splitlines():
            rec = {}
            for f in 'url', 'pushurl':
                try:
                    v = self._run_git('config remote.%s.%s' % (remote, f)
                                      , expect_fail=True
                                      #, expect_stderr=True
                                     )
                    if v is not None:
                        rec[f] = v
                except CommandError:
                    # must have no value
                    pass
            if remote in containing_remotes:
                rec['contains'] = True
            remotes[remote] = rec
        return remotes

    @property
    def tracked_remote(self):
        branch = self.branch
        if not branch:
            return None
        return self._run_git(
                'config branch.%s.remote' % (branch,)
                #, expect_stderr=True
                , expect_fail=True
        ) or None         # want explicit None

    @property
    def branch(self):
        if self._branch is None:
            try:
                branch = self._run_git('symbolic-ref --short HEAD')
            except CommandError:
                # We're in a detached state.
                return None
            self._branch = branch
        return self._branch

    def has_revision(self, revision):
        """Does the repository have `revision`?
        """
        out = self._run_git(
            ["rev-parse", "--quiet", "--verify", revision + "^{commit}"],
            expect_fail=True)
        return out is not None


class VCSTracer(DistributionTracer):
    """Resolve files into VCS repositories they are contained with

    TODO: generalize to other common VCS (svn, hg, bzr) which should win one
    over the other depending on the path, e.g. for path

    Devel notes:
    - Whenever we allow for some "hypothetical" ownership, /a/b/c/d/f with
      git at /a/b and hg at /a/b/c/ it must be hg repo which contains it.
      But may be we wouldn't need that and should just grab all the repositories
      for which we have strict membership hits, and then all possible which sit
      above files which were not found belonging to any -- that is where we
      could go with the most nested one and include it as well, although without
      assigning any file specifically to it
    - CVS and svn might be the first ones to consider since they have corresponding
      .svn and CVS at every level in their hierarchy so easier to catch

    Assumptions:
    - VCS operate at the "directories" level, i.e. the file /a/b/c can't be contained
      in one VCS whenever /a/b/d file in another
    """

    SHIMS = (SVNRepoShim, GitRepoShim)

    def _init(self):
        # dictionary to contain per each inspected/known directory a VCS
        # instance it belongs to
        self._known_repos = {}

    def identify_distributions(self, files):
        repos, remaining_files = self.identify_packages_from_files(
            files, root_key="path")
        pkgs_per_distr = defaultdict(list)
        for repo in repos:
            pkgs_per_distr[repo._distribution].append(repo)
        for dist_class, repos in pkgs_per_distr.items():
            yield dist_class(name=dist_class._cmd,
                             packages=repos), remaining_files

    def _get_packagefields_for_files(self, files):
        out = {}
        for f in files:
            lgr.log(6, "%s testing file %s", self, f)
            shim = self._resolve_file(f)
            if not shim:
                continue
            # we probably do not want all the attributes to just report which
            # repo it is, so let's report path and type
            # out[f] = dict(
            #     (a.name, getattr(shim, a.name)) for a in shim._vcs_class.__attrs_attrs__
            #     if a.name not in {'files'}
            # )
            out[f] = {
                'path': shim.path,
                # 'repo_class': shim._vcs_class,
            }
            # the rest of the attrs will be taken by using _known_repos
            # in _create_package
        return out

    def _create_package(self, path):
        # TODO:  we might want to mark those which are found to belong to pkg
        #  files which are dirty.
        shim = self._known_repos[path]
        attrs = dict(
            (a.name, getattr(shim, a.name)) for a in shim._vcs_class.__attrs_attrs__
            if a.name not in {'files'}  # those will be populated later
        )
        attrs = only_with_values(attrs)
        return instantiate_attr_object(shim._vcs_class, attrs)

    def _resolve_file(self, path):
        """Given a path, return path of the repository it belongs to"""
        # very naive just to get a ball rolling
        if not isabs(path):
            path = abspath(path)
        dirpath = path if isdir(path) else dirname(path)

        # quick check first
        if dirpath in self._known_repos:
            return self._known_repos[dirpath]

        # it could still be a subdirectory known to the repository known above it
        for repo in self._known_repos.values():
            # XXX this design is nohow accounts for some fancy cases where
            # someone could use GIT_TREE and other trickery to have out of the
            # directory checkout.  May be some time we would get there but
            # for now
            # should be ok

            # could be less efficient than some str manipulations but ok for now
            if os.path.commonprefix((repo.path, path)) != repo.path:
                continue

            # could be less efficient than some str manipulations but ok for now
            # since we rely on a strict check (must be registered within the repo)
            if repo.owns_path(path):
                return repo

        # ok -- if it is not among known repos, we need to 'sniff' around
        # if there is a repository at that path
        for Shim in self.SHIMS:
            lgr.log(5, "Trying %s for path %s", Shim, path)
            shim = Shim.get_at_dirpath(self._session, dirpath) \
                if lexists(dirpath) else None
            if shim:
                # so there is one nearby -- record it
                self._known_repos[shim.path] = shim
                # but it might still not to know about the file
                if shim.owns_path(path):
                    return shim
                # if not -- just keep going to the next candidate repository
        return None



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NICEMAN package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test functionality of example.py

"""

from niceman.examples.example import get_url_for_packages

__docformat__ = 'restructuredtext'

# from niceman.examples.example import
from niceman.tests.utils import assert_equal


def test_get_url_for_packages():
    assert_equal(get_url_for_packages({'cmtk': '3.2'}), {'cmtk': 'http://example.com/cmtk_3.2.deb'})
    assert_equal(get_url_for_packages(
        {'cmtk': '3.1'}), {'cmtk': 'http://example.com/cmtk_3.1.deb'})


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Common tests configuration for resources
"""

from appdirs import AppDirs
import os
import pytest
import shutil


@pytest.fixture(scope="module")
def resource_test_dir():
	dirs = AppDirs('niceman')
	test_dir = os.path.join(dirs.user_cache_dir, 'resource_test')
	if not os.path.exists(test_dir):
		os.makedirs(test_dir)
	yield test_dir
	shutil.rmtree(test_dir)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Common interface options

"""

__docformat__ = 'restructuredtext'

from niceman.support.param import Parameter
from niceman.support.constraints import EnsureChoice
from niceman.support.constraints import EnsureInt, EnsureNone, EnsureStr


trace_opt = Parameter(
    args=("--trace",),
    action="store_true",
    doc="""if set, trace execution within the environment""")


#
# Resource specifications
#

resref_arg = Parameter(
    args=("resref",),
    metavar="RESOURCE",
    doc="""Name or ID of the resource to operate on. To see available resources, run
    'niceman ls'""",
    constraints=EnsureStr() | EnsureNone())

resref_opt = Parameter(
    args=("-r", "--resource",),
    dest="resref",
    metavar="RESOURCE",
    doc="""Name or ID of the resource to operate on. To see available resources, run
    'niceman ls'""",
    constraints=EnsureStr() | EnsureNone())

resref_type_opt = Parameter(
    args=("--resref-type",),
    metavar="TYPE",
    doc="""A resource can be referenced by its name or ID.  In the unlikely
    case that a name collides with an ID, explicitly specify 'name' or 'id' to
    disambiguate.""",
    constraints=EnsureChoice("auto", "name", "id"))



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NICEMAN package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Progress bar implementations to be used.

Should not be imported until we know that interface needs it
"""

import sys

#
# Haven't found an ideal progress bar yet, so to make things modular etc
# we will provide our interface and adapters for few popular ones
#

class ProgressBarBase(object):
    def __init__(self):
        self._prev_value = 0

    def update(self, size, increment=False):
        if increment:
            self._prev_value += size
        else:
            self._prev_value = size

    def start(self):
        pass

    def finish(self):
        self._prev_value = 0

progressbars = {}
try:
    from progressbar import Bar, ETA, FileTransferSpeed, \
        Percentage, ProgressBar, RotatingMarker

    # TODO: might better delegate to an arbitrary bar?
    class BarWithFillText(Bar):
        """A progress bar widget which fills the bar with the target text"""

        def __init__(self, fill_text, **kwargs):
            super(BarWithFillText, self).__init__(**kwargs)
            self.fill_text = fill_text

        def update(self, pbar, width):
            orig = super(BarWithFillText, self).update(pbar, width)
            # replace beginning with the title
            if len(self.fill_text) > width:
                # TODO:  make it fancier! That we also at the same time scroll it from
                # the left so it does end up at the end with the tail but starts with
                # the beginning
                fill_text = '...' + self.fill_text[-(width-4):]
            else:
                fill_text = self.fill_text
            fill_text = fill_text[:min(len(fill_text), int(round(width * pbar.percentage()/100.)))]
            return fill_text + " " + orig[len(fill_text)+1:]

    class progressbarProgressBar(ProgressBarBase):
        """Adapter for progressbar.ProgressBar"""

        backend = 'progressbar'

        def __init__(self, label='', fill_text=None, maxval=None, unit='B', out=sys.stdout):
            super(progressbarProgressBar, self).__init__()
            assert(unit == 'B')  # none other "supported" ATM
            bar = dict(marker=RotatingMarker())
            # TODO: RF entire messaging to be able to support multiple progressbars at once
            widgets = ['%s: ' % label,
                       BarWithFillText(fill_text=fill_text, marker=RotatingMarker()), ' ',
                       Percentage(), ' ',
                       ETA(), ' ',
                       FileTransferSpeed()]
            self._pbar = ProgressBar(widgets=widgets, maxval=maxval, fd=out).start()

        def update(self, size, increment=False):
            self._pbar.update(self._prev_value + size if increment else size)
            super(progressbarProgressBar, self).update(size, increment=increment)

        def start(self):
            super(progressbarProgressBar, self).start()
            self._pbar.start()

        def finish(self):
            if self._pbar:
                self._pbar.finish()
            super(progressbarProgressBar, self).finish()

    progressbars['progressbar'] = progressbarProgressBar
except ImportError:  # pragma: no cover
    pass

try:
    from tqdm import tqdm

    class tqdmProgressBar(ProgressBarBase):
        """Adapter for tqdm.ProgressBar"""

        backend = 'tqdm'

        def __init__(self, label='', fill_text=None, maxval=None, unit='B', out=sys.stdout):
            super(tqdmProgressBar, self).__init__()
            self._pbar_params = dict(desc=label, unit=unit, unit_scale=True, total=maxval, file=out)
            self._pbar = None

        def _create(self):
            if self._pbar is None:
                # set miniters to 1, and mininterval to 0, so it always updates
                self._pbar = tqdm(miniters=1, mininterval=0, **self._pbar_params)

        def update(self, size, increment=False):
            self._create()
            inc = size - self._prev_value
            self._pbar.update(size if increment else inc)
            super(tqdmProgressBar, self).update(size, increment=increment)

        def start(self):
            super(tqdmProgressBar, self).start()
            self._create()

        def finish(self):
            # be tollerant to bugs in those
            try:
                if self._pbar is not None:
                    self._pbar.close()
            finally:
                self._pbar = None
            try:
                super(tqdmProgressBar, self).finish()
            except Exception as exc:  # pragma: no cover
                #lgr.debug("Finishing tqdmProgresBar thrown %s", str_exc(exc))
                pass

    progressbars['tqdm'] = tqdmProgressBar
except ImportError:  # pragma: no cover
    pass

assert len(progressbars), "We need tqdm or progressbar library to report progress"




# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from os.path import join as opj, dirname
from niceman.cmdline.main import main

from niceman.utils import swallow_outputs, swallow_logs
from niceman.tests.utils import assert_in, assert_not_in, assert_equal
from niceman.tests.utils import assert_in_in

from pytest import raises

diff_1_yaml = opj(dirname(__file__), 'files', 'diff_1.yaml')
diff_2_yaml = opj(dirname(__file__), 'files', 'diff_2.yaml')
empty_yaml = opj(dirname(__file__), 'files', 'empty.yaml')

multi_debian_yaml = opj(dirname(__file__), 'files', 'multi_debian.yaml')

def test_multi_debian_files():
    with swallow_logs() as log:
        args = ['diff', multi_debian_yaml, diff_1_yaml]
        with raises(SystemExit):
            main(args)
        assert_in_in("multiple <class 'niceman.distributions.debian.DebianDistribution'> found", log.lines)
    with swallow_logs() as log:
        args = ['diff', diff_1_yaml, multi_debian_yaml]
        with raises(SystemExit):
            main(args)
        assert_in_in("multiple <class 'niceman.distributions.debian.DebianDistribution'> found", log.lines)


def test_same():
    with swallow_outputs() as outputs:
        args = ['diff', diff_1_yaml, diff_1_yaml]
        main(args)
        assert_equal(outputs.out, '')
        assert_equal(outputs.err, '')


def test_diff_files():
    with swallow_outputs() as outputs:
        args = ['diff', diff_1_yaml, diff_2_yaml]
        main(args)
        assert_equal(outputs.err, '')
        assert_in('Files:', outputs.out)
        assert_in('< /etc/a', outputs.out)
        assert_in('> /etc/c', outputs.out)
        assert_not_in('< /etc/b', outputs.out)
        assert_not_in('> /etc/b', outputs.out)


def test_diff_debian_packages():
    with swallow_outputs() as outputs:
        args = ['diff', diff_1_yaml, diff_2_yaml]
        main(args)
        assert_equal(outputs.err, '')
        assert_in('Debian packages:', outputs.out)
        assert_in('< lib1only x86 2:1.6.4-3', outputs.out)
        assert_in('> lib2only x86 2:1.6.4-3', outputs.out)
        assert_in('< libarchdiff x86 2.4.6', outputs.out)
        assert_in('> libarchdiff amd64 2.4.6', outputs.out)
        assert_in('Debian package libversdiff x86:', outputs.out)
        assert_in('< 2.4.6', outputs.out)
        assert_in('> 2.4.7', outputs.out)
        assert_not_in('libsame', outputs.out)


def test_diff_conda_packages():
    with swallow_outputs() as outputs:
        args = ['diff', diff_1_yaml, diff_2_yaml]
        main(args)
        assert_equal(outputs.err, '')
        assert_in('Conda packages:', outputs.out)
        assert_in('< c_lib1only py36_0 2:1.6.4-3', outputs.out)
        assert_in('> c_lib2only py36_0 2:1.6.4-3', outputs.out)
        assert_in('< c_libbuilddiff py36_0 2.4.6', outputs.out)
        assert_in('> c_libbuilddiff hdf63c60_3 2.4.6', outputs.out)
        # TO DO: ensure the version strings (second and third lines below) 
        # come from the conda report -- these could just match the debian 
        # output checked in test_diff_debian_packages()
        assert_in('Conda package c_libversdiff py36_0:', outputs.out)
        assert_in('< 2.4.6', outputs.out)
        assert_in('> 2.4.7', outputs.out)
        assert_not_in('c_libsame', outputs.out)


def test_diff_no_distributions():
    with swallow_outputs() as outputs:
        args = ['diff', diff_1_yaml, empty_yaml]
        main(args)
        assert_equal(outputs.err, '')
        assert_in('Debian packages:', outputs.out)
        assert_in('< lib1only x86 2:1.6.4-3', outputs.out)
        assert_in('< libsame x86 2.4.6', outputs.out)
        assert_in('< libarchdiff x86 2.4.6', outputs.out)
        assert_in('< libversdiff x86 2.4.6', outputs.out)
        assert_in('Conda packages:', outputs.out)
        assert_in('< c_lib1only py36_0 2:1.6.4-3', outputs.out)
        assert_in('< c_libsame py36_0 2.4.6', outputs.out)
        assert_in('< c_libbuilddiff py36_0 2.4.6', outputs.out)
        assert_in('< c_libversdiff py36_0 2.4.6', outputs.out)
        assert_in('Files:', outputs.out)
        assert_in('< /etc/a', outputs.out)
        assert_in('< /etc/b', outputs.out)


def test_diff_git():
    with swallow_outputs() as outputs:
        args = ['diff', diff_1_yaml, diff_2_yaml]
        main(args)
        assert_equal(outputs.err, '')
        assert_in('Git repositories:', outputs.out)
        assert_in('< 43e8e6577c7bf493ddb01ea7d49bef7dc7a6643b (/path/to/git/repo/1/only)', outputs.out)
        assert_in('> 64b1865267891fdd1a45251ca6f32df213dc546e (/path/to/git/repo/2/only)', outputs.out)
        assert_in('Git repository 5b8267181f6cae8dc37aeef21ea54171bd932522', outputs.out)
        assert_in('< 3e3aaa73a9c0ca061c7679af5fa7318e70f528ac (/path/1/to/different/git/commit)', outputs.out)
        assert_in('> 9d199f7fa7e6f691719e0860c5cf81193e815ad5 (/path/2/to/different/git/commit)', outputs.out)
        assert_not_in('/path/1/to/common/git/repo', outputs.out)
        assert_not_in('/path/2/to/common/git/repo', outputs.out)
        assert_not_in('99ac7f69a070077038a9eb9eca61c028db97181d', outputs.out)
        assert_not_in('d057b128759d80a47500adba0c4d3e95092bb87f', outputs.out)


def test_diff_svn():
    with swallow_outputs() as outputs:
        args = ['diff', diff_1_yaml, diff_2_yaml]
        main(args)
        assert_equal(outputs.err, '')
        assert_in('SVN repositories:', outputs.out)
        assert_in('< c8ed47ab-45c9-818d-5d62-549dcc6d97d4 (/path/to/svn/repo/1/only)', outputs.out)
        assert_in('> d7192e3a-60de-5caa-ccdc9525dea75aabf (/path/to/svn/repo/2/only)', outputs.out)
        assert_in('SVN repository 95e4b738-84c7-154c-f082-34d40e21fdd4', outputs.out)
        assert_in('< 12 (/path/1/to/different/svn/commit)', outputs.out)
        assert_in('> 14 (/path/2/to/different/svn/commit)', outputs.out)
        assert_not_in('/path/1/to/common/svn/repo', outputs.out)
        assert_not_in('/path/2/to/common/svn/repo', outputs.out)
